# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import itertools
import os
from collections import defaultdict
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, Mapping, Optional, Protocol, TypeVar

import torch
from torch import nn

from nemo.lightning.megatron_init import initialize_model_parallel_for_nemo
from nemo.utils import logging

NEMO_MEGATRON_MODEL_PARALLEL_APPSTATE_OVERRIDE = "NEMO_MEGATRON_MODEL_PARALLEL_APPSTATE_OVERRIDE"


if TYPE_CHECKING:
    from lightning.fabric.utilities.types import Optimizable
    from megatron.core.model_parallel_config import ModelParallelConfig


class SharedStateDictProtocol(Protocol):
    """ """

    def sharded_state_dict(self, prefix="", metadata: Optional[dict] = None):
        """ """
        ...


def init_parallel_ranks(
    world_size: int,
    global_rank: int,
    local_rank: int,
    parallel_config: "ModelParallelConfig",
    seed=1234,
    fp8=False,
) -> None:
    """
    Initializes the parallel ranks for distributed training.

    This function sets up the parallel ranks based on the provided world size, global rank, local rank,
    and parallel configuration. It also sets the seed for random number generation and determines whether
    to use fp8 precision.

    Args:
        world_size (int): The total number of processes participating in the distributed training.
        global_rank (int): The rank of the current process in the distributed training setup.
        local_rank (int): The rank of the current process within its machine.
        parallel_config (ModelParallelConfig): The configuration object containing settings for model parallelism.
        seed (int, optional): The seed for random number generation. Defaults to 1234.
        fp8 (bool, optional): Whether to use fp8 precision for model parameters. Defaults to False.
    """
    from nemo.utils import AppState

    app_state = AppState()

    if os.environ.get(NEMO_MEGATRON_MODEL_PARALLEL_APPSTATE_OVERRIDE, "false").lower() == "true":
        init_world_size = app_state.tensor_model_parallel_size * app_state.pipeline_model_parallel_size
        init_global_rank = app_state.global_rank
        init_local_rank = app_state.local_rank
    else:
        init_world_size = world_size
        pp = parallel_config.pipeline_model_parallel_size or 1
        if world_size < pp:
            raise ValueError(f"Expected world_size ({world_size}) to be greater than/equal to pipeline size ({pp})")
        init_global_rank = global_rank
        init_local_rank = local_rank

    initialize_model_parallel_for_nemo(
        world_size=init_world_size,
        global_rank=init_global_rank,
        local_rank=init_local_rank,
        tensor_model_parallel_size=parallel_config.tensor_model_parallel_size,
        expert_model_parallel_size=parallel_config.expert_model_parallel_size,
        expert_tensor_parallel_size=parallel_config.expert_tensor_parallel_size,
        pipeline_model_parallel_size=parallel_config.pipeline_model_parallel_size,
        pipeline_model_parallel_comm_backend=parallel_config.pipeline_model_parallel_comm_backend,
        virtual_pipeline_model_parallel_size=parallel_config.virtual_pipeline_model_parallel_size,
        context_parallel_size=parallel_config.context_parallel_size,
        seed=seed,
        use_fp8=fp8,
        init_mpi_proc_group=getattr(parallel_config, "tp_comm_overlap", False)
        and getattr(parallel_config, "tp_comm_bootstrap_backend", None) == 'mpi',
        use_te_rng_tracker=getattr(parallel_config, "use_te_rng_tracker", False),
        use_sharp=getattr(parallel_config, "use_sharp", False),
        create_all_gather_group=getattr(parallel_config, "create_all_gather_group", False),
        use_tp_pp_dp_mapping=getattr(parallel_config, "use_tp_pp_dp_mapping", False),
        num_distributed_optimizer_instances=getattr(parallel_config, "num_distributed_optimizer_instances", 1),
        nccl_communicator_config_path=getattr(parallel_config, "nccl_communicator_config_path", None),
        use_gloo_process_groups=getattr(parallel_config, "use_gloo_process_groups", True),
        # apex_transformer_log_level=self.cfg.get('apex_transformer_log_level', 30),
    )


def init_model_parallel(model: Optional[nn.Module] = None) -> None:
    """Initializes Megatron-LM model parallel if using model parallelism."""
    import inspect
    import torch.distributed
    from megatron.core import parallel_state

    from nemo.utils import AppState

    app_state = AppState()

    # we initialize megatron-lm model parallel and data parallel groups
    # after initializing DDP with PTL.
    if app_state.model_parallel_size is not None:
        # destroy groups in case they have already been created
        # this happens with multiple calls to trainer.test for example
        parallel_state.destroy_model_parallel()
        if torch.distributed.is_initialized():
            sig = inspect.signature(parallel_state.initialize_model_parallel)

            # Check if Megatron-LM supports create_all_gather_group parameter (backward compatibility)
            extra_kwargs = {}
            if 'create_all_gather_group' in sig.parameters:
                extra_kwargs['create_all_gather_group'] = app_state.create_all_gather_group

            parallel_state.initialize_model_parallel(
                tensor_model_parallel_size=app_state.tensor_model_parallel_size,
                pipeline_model_parallel_size=app_state.pipeline_model_parallel_size,
                virtual_pipeline_model_parallel_size=app_state.virtual_pipeline_model_parallel_size,
                pipeline_model_parallel_comm_backend=app_state.pipeline_model_parallel_comm_backend,
                context_parallel_size=app_state.context_parallel_size,
                expert_model_parallel_size=app_state.expert_model_parallel_size,
                expert_tensor_parallel_size=app_state.expert_tensor_parallel_size,
                use_sharp=app_state.use_sharp,
                order="tp-cp-ep-pp-dp" if app_state.use_tp_pp_dp_mapping else "tp-cp-ep-dp-pp",
                num_distributed_optimizer_instances=app_state.num_distributed_optimizer_instances,
                nccl_communicator_config_path=app_state.nccl_communicator_config_path,
                create_gloo_process_groups=app_state.use_gloo_process_groups,
                **extra_kwargs,
            )

            # assert that fake tp and pp rank match after model parallel init
            assert app_state.tensor_model_parallel_rank == parallel_state.get_tensor_model_parallel_rank()
            assert app_state.pipeline_model_parallel_rank == parallel_state.get_pipeline_model_parallel_rank()
            assert app_state.expert_tensor_parallel_rank == parallel_state.get_expert_tensor_parallel_rank()

            app_state.tensor_model_parallel_group = parallel_state.get_tensor_model_parallel_group()
            app_state.data_parallel_group = parallel_state.get_data_parallel_group()
            app_state.data_parallel_rank = parallel_state.get_data_parallel_rank()
            app_state.data_parallel_size = parallel_state.get_data_parallel_world_size()
            app_state.pipeline_model_parallel_group = parallel_state.get_pipeline_model_parallel_group()

            # create MPI process group for UCX-based communication APIs
            if app_state.init_mpi_proc_group:
                torch.distributed.new_group(backend="mpi")


def set_model_parallel_attributes(model, parallelism):
    """ """
    # Right now mcore sub-classes ModelParellelConfig, we should remove that
    # Given Lightning's structure it would be better if parallelism is a different object
    # Since then it can be passed to the Strategy
    # Note: Importing nemo.lightning.pytorch.strategies creates an import cycle.
    from megatron.core.transformer.transformer_config import TransformerConfig

    has_mcore_config = isinstance(getattr(model, "config", None), TransformerConfig)
    if has_mcore_config and hasattr(model, "configure_model"):
        config: TransformerConfig = model.config
        for attr_name in filter(lambda x: not x.startswith('__'), dir(parallelism)):
            if not hasattr(config, attr_name):
                continue
            setattr(config, attr_name, getattr(parallelism, attr_name))
            if hasattr(config, "__io__"):
                setattr(config.__io__, attr_name, getattr(parallelism, attr_name))
        if hasattr(config, '__post_init__'):
            # MCore does not use args in __post_init__
            # @akoumparouli: is there a better way (e.g. reinit config)?
            config.__post_init__()

        return config

    return None


@contextmanager
def megatron_lazy_init_context(config) -> Generator[None, None, None]:
    """ """
    try:
        from megatron.core.extensions import transformer_engine as _te

        original = _te._get_extra_te_kwargs  # noqa: SLF001

        def _get_extra_te_kwargs_meta(c):
            """Forces device to meta"""
            kwargs = original(c)
            kwargs['device'] = 'meta'
            return kwargs

        _te._get_extra_te_kwargs = _get_extra_te_kwargs_meta  # noqa: SLF001
    except ImportError:
        pass

    _orig_perform_initialization = config.perform_initialization
    _orig_use_cpu_initialization = config.use_cpu_initialization

    config.perform_initialization = False
    config.use_cpu_initialization = True

    yield

    try:
        from megatron.core.extensions import transformer_engine as _te

        _te._get_extra_te_kwargs = original  # noqa: SLF001
    except ImportError:
        pass

    config.perform_initialization = _orig_perform_initialization
    config.use_cpu_initialization = _orig_use_cpu_initialization


@contextmanager
def megatron_cpu_init_context(config) -> Generator[None, None, None]:
    """ """
    _orig_use_cpu_initialization = config.use_cpu_initialization

    config.use_cpu_initialization = True

    yield

    config.use_cpu_initialization = _orig_use_cpu_initialization


ModelT = TypeVar("ModelT", bound=nn.Module)


class GradScaler(torch.cuda.amp.GradScaler):
    """
    Gradient sclaer for model-parallel inf check. The inf in gradients are checked across tensor-parallel
    ranks in (1) executing optimizer step and (2) gradient scaler update.

    """

    def __init__(
        self,
        init_scale=2.0**16,
        growth_factor=2.0,
        backoff_factor=0.5,
        growth_interval=2000,
        enabled=True,
        hysteresis=1,
    ):
        super().__init__(
            init_scale=init_scale,
            growth_factor=growth_factor,
            backoff_factor=backoff_factor,
            growth_interval=growth_interval,
            enabled=enabled,
        )
        self.optimizer_update_skipped: Optional[bool] = None
        self.hysteresis = hysteresis
        self._hysteresis_tracker = self.hysteresis

    def _unscale_grads_(self, optimizer, *args):
        if getattr(optimizer, "_custom_amp_unscale_grads", False):
            return optimizer.unscale_grads(*args)
        else:
            return super()._unscale_grads_(optimizer, *args)

    def _maybe_opt_step(self, optimizer, optimizer_state, *args, **kwargs):
        from megatron.core import parallel_state

        retval = None
        found_inf = torch.cuda.FloatTensor([sum(v.item() for v in optimizer_state["found_inf_per_device"].values())])

        # Update across all model parallel instances.
        torch.distributed.all_reduce(
            found_inf,
            op=torch.distributed.ReduceOp.MAX,
            group=parallel_state.get_model_parallel_group(),
        )

        if found_inf.item() == 0:
            retval = optimizer.step(*args, **kwargs)
            self.optimizer_update_skipped = False
        else:
            self.optimizer_update_skipped = True
        return retval

    def update(self, new_scale=None):
        """
        Updates to native grad scaler update function.
        1. Check inf across model-parallel ranks.
        2. Update hysteresis tracker.
        3. Apply hysteresis to grad scale update.
        """
        from megatron.core import parallel_state

        if not self._enabled:
            return

        _scale, _growth_tracker = self._check_scale_growth_tracker("update")

        if new_scale is not None:
            # Accept a new user-defined scale.
            if isinstance(new_scale, float):
                self._scale.fill_(new_scale)  # type: ignore[union-attr]
            else:
                reason = (
                    "new_scale should be a float or a 1-element torch.cuda.FloatTensor with" " requires_grad=False."
                )
                assert isinstance(new_scale, torch.cuda.FloatTensor), reason  # type: ignore[attr-defined]
                assert new_scale.numel() == 1, reason
                assert new_scale.requires_grad is False, reason
                self._scale.copy_(new_scale)  # type: ignore[union-attr]
        else:
            # Consume shared inf/nan data collected from optimizers to update the scale.
            # If all found_inf tensors are on the same device as self._scale, this operation is asynchronous.
            found_infs = [
                found_inf.to(device=_scale.device, non_blocking=True)
                for state in self._per_optimizer_states.values()
                for found_inf in state["found_inf_per_device"].values()
            ]

            assert len(found_infs) > 0, "No inf checks were recorded prior to update."

            found_inf_combined = found_infs[0]

            # Update across all model parallel instances.
            torch.distributed.all_reduce(
                found_inf_combined,
                op=torch.distributed.ReduceOp.MAX,
                group=parallel_state.get_model_parallel_group(),
            )

            if len(found_infs) > 1:
                for i in range(1, len(found_infs)):
                    found_inf = found_infs[i]
                    # Update across all model parallel instances.
                    torch.distributed.all_reduce(
                        found_inf,
                        op=torch.distributed.ReduceOp.MAX,
                        group=parallel_state.get_model_parallel_group(),
                    )
                    found_inf_combined += found_inf

            if found_inf_combined > 0:
                self._hysteresis_tracker -= 1
                if self._hysteresis_tracker <= 0:
                    # When hysteresis becomes zero, follow the native grad scale update rule.
                    # Increase scale and reset growth tracker
                    torch._amp_update_scale_(  # noqa: SLF001
                        _scale,
                        _growth_tracker,
                        found_inf_combined,
                        self._growth_factor,
                        self._backoff_factor,
                        self._growth_interval,
                    )
                else:
                    # Only reset the growth tracker when hysteresis is larger than zero
                    _growth_tracker.fill_(0.0)
            else:
                # When no inf found, follow the native grad scale update rule.
                # Increment growth_tracker, update scale when growth tracker reaches the interval, and
                # reset the hysteresis tracker.
                torch._amp_update_scale_(  # noqa: SLF001
                    _scale,
                    _growth_tracker,
                    found_inf_combined,
                    self._growth_factor,
                    self._backoff_factor,
                    self._growth_interval,
                )
                self._hysteresis_tracker = self.hysteresis

        # To prepare for next iteration, clear the data collected from optimizers this iteration.
        self._per_optimizer_states = defaultdict(
            torch.cuda.amp.grad_scaler._refresh_per_optimizer_state  # noqa: SLF001
        )

    def state_dict(self):
        """
        Add hysteresis_tracker to the native functions' state_dict.
        """
        return (
            {
                "scale": self.get_scale(),
                "growth_factor": self._growth_factor,
                "backoff_factor": self._backoff_factor,
                "growth_interval": self._growth_interval,
                "_growth_tracker": self._get_growth_tracker(),
                "_hysteresis_tracker": self._hysteresis_tracker,
            }
            if self._enabled
            else {}
        )

    def load_state_dict(self, state_dict):
        """
        Load hysteresis_tracker in addition to the state dict of the native function.
        """
        if not self._enabled:
            return

        if len(state_dict) == 0:
            raise RuntimeError(
                "The source state dict is empty, possibly because it was saved "
                "from a disabled instance of GradScaler."
            )

        self._init_scale = state_dict["scale"]
        if self._scale is not None:
            self._scale.fill_(state_dict["scale"])
        self._growth_factor = state_dict["growth_factor"]
        self._backoff_factor = state_dict["backoff_factor"]
        self._growth_interval = state_dict["growth_interval"]
        self._init_growth_tracker = state_dict["_growth_tracker"]
        if self._growth_tracker is not None:
            self._growth_tracker.fill_(state_dict["_growth_tracker"])
        if "_hysterisis_tracker" in state_dict:
            self._hysteresis_tracker = state_dict["_hysterisis_tracker"]
        else:
            self._hysteresis_tracker = 1


def enable_nvidia_optimizations() -> None:
    """These optimizations are present in NVIDIA NGC PyTorch Containers."""
    # NVIDIA container version check
    nvidia_torch_version = os.getenv("NVIDIA_PYTORCH_VERSION", None)
    if nvidia_torch_version is not None:
        try:
            NVIDIA_TORCH_MAJOR = int(nvidia_torch_version.split(".")[0])
        except Exception:
            NVIDIA_TORCH_MAJOR = 0
        try:
            NVIDIA_TORCH_MINOR = int(nvidia_torch_version.split(".")[1])
        except Exception:
            NVIDIA_TORCH_MINOR = 0

        # NVFUSER available starting with 21.11
        if NVIDIA_TORCH_MAJOR >= 21 or (NVIDIA_TORCH_MAJOR == 21 and NVIDIA_TORCH_MINOR >= 11):
            # NVFUSER
            torch._C._jit_set_profiling_executor(True)  # noqa: SLF001
            torch._C._jit_set_profiling_mode(True)  # noqa: SLF001
            torch._C._jit_override_can_fuse_on_cpu(False)  # noqa: SLF001
            torch._C._jit_override_can_fuse_on_gpu(False)  # noqa: SLF001
            torch._C._jit_set_texpr_fuser_enabled(False)  # noqa: SLF001
            # torch._C._jit_set_nvfuser_enabled(True)
            torch._C._debug_set_autodiff_subgraph_inlining(False)  # noqa: SLF001
    else:
        # Not a Nvidia container. NVFUSER Dependency check is on users
        pass


def optimizer_sharded_state_dict(
    model: SharedStateDictProtocol,
    optimizer: "Optimizable",
    is_loading: bool = False,
    sharding_type: Optional[str] = None,
    metadata: Optional[dict] = None,
) -> Dict[str, torch.Tensor]:
    """
    Sharded state dictionary for an MainParamsOptimizerWrapper.
    Used to save and load the optimizer state when training with distributed_checkpoint.

    Args:
        model (SharedStateDictProtocol): model with a `sharded_state_dict` method
        optimizer (Optimizable): optimizer to get the state dict of
        is_loading (bool, optional): set to True if the sharded state dict is intended
            for checkpoint loading (as opposed to saving). Defaults to False.
        sharding_type (str, optional): deprecated, use metadata flags instead.
        metadata (dict, optional): sharded state dict metadata passed from the framework.
            Used to control the details of sharded state dict creation, in particular
            the state dict format of the DistributedOptimizer with the flag
            `distrib_optim_sharding_type`. Defaults to None (empty metadata).

    Returns
    -------
        dict: The sharded state dictionary for the optimizer
    Raises:
        ValueError: If a parameter ID does not match any model sharded parameter.
    """
    from megatron.core.dist_checkpointing.optimizer import (
        get_param_id_to_sharded_param_map,
        make_sharded_optimizer_tensor,
        optim_state_to_sharding_state,
    )

    from nemo.core.optim import MainParamsOptimizerWrapper
    from nemo.core.optim.optimizers import init_optimizer_states

    model_sharded_state_dict = model.sharded_state_dict(metadata=metadata)

    # remove _extra_state
    model_sharded_state_dict = {
        key: value for key, value in model_sharded_state_dict.items() if not key.endswith("_extra_state")
    }

    if sharding_type is not None:
        logging.warning("sharding_type is deprecated, please use `metadata['distrib_optim_sharding_type']` instead")
        if metadata is None:
            metadata = {}
        if 'distrib_optim_sharding_type' not in metadata:
            metadata["distrib_optim_sharding_type"] = sharding_type

    if hasattr(optimizer, "sharded_state_dict"):
        return optimizer.sharded_state_dict(
            model_sharded_state_dict,
            is_loading=is_loading,
            metadata=metadata,
        )

    if not isinstance(optimizer, MainParamsOptimizerWrapper):
        # Regular optimizer, e.g. Adam or FusedAdam
        init_optimizer_states(optimizer)
        optimizer_state_dict = optimizer.state_dict()
        id_to_sharded_param_map = get_param_id_to_sharded_param_map(
            model_sharded_state_dict=model_sharded_state_dict,
            optim_params_iter=itertools.chain.from_iterable(g['params'] for g in optimizer.param_groups),
        )
        optim_state_to_sharding_state(optimizer_state_dict, id_to_sharded_param_map)
        return optimizer_state_dict

    optimizer_state_dict: Dict[str, Any] = optimizer.state_dict()

    id_to_sharded_param_map = get_param_id_to_sharded_param_map(
        model_sharded_state_dict=model_sharded_state_dict,
        optim_params_iter=itertools.chain.from_iterable(g for g in optimizer.float16_groups),
    )

    # Convert fp32_from_fp16_params
    assert len(optimizer_state_dict["fp32_from_fp16_params"]) == len(optimizer_state_dict["optimizer"]["param_groups"])

    def get_safe(param_id):
        try:
            return id_to_sharded_param_map[param_id]
        except KeyError as e:
            raise ValueError(f"Param id {param_id} does not match any model sharded param") from e

    optimizer_state_dict["fp32_from_fp16_params"] = [
        [
            make_sharded_optimizer_tensor(get_safe(param_id), fp32_param, prefix="optimizer.state.fp32_param")
            for param_id, fp32_param in zip(state_group["params"], fp32_group)
        ]
        for fp32_group, state_group in zip(
            optimizer_state_dict["fp32_from_fp16_params"],
            optimizer_state_dict["optimizer"]["param_groups"],
        )
    ]

    # Convert state
    optim_state_to_sharding_state(optimizer_state_dict["optimizer"], id_to_sharded_param_map)

    return optimizer_state_dict


def load_model_state_dict(megatron_parallel, checkpoint: Mapping[str, Any], strict: bool = True) -> None:
    """ """
    from megatron.core import parallel_state
    from megatron.core.dist_checkpointing.validation import StrictHandling, parse_strict_flag

    # convert from StrictHandling to bool for PTL
    if strict is not None and not isinstance(strict, bool):
        strict = parse_strict_flag(strict)
        strict_options = [
            StrictHandling.ASSUME_OK_UNEXPECTED,
            StrictHandling.RAISE_UNEXPECTED,
            StrictHandling.RAISE_ALL,
        ]
        strict = strict in strict_options

    try:
        from megatron.core.distributed.custom_fsdp import FullyShardedDataParallel

        have_custom_fsdp = True
    except ImportError or ModuleNotFoundError:
        have_custom_fsdp = False

    try:
        from megatron.core.distributed import FullyShardedDataParallel

        have_megatron_fsdp = True
    except ImportError or ModuleNotFoundError:
        have_megatron_fsdp = False

    for index, module in enumerate(megatron_parallel):
        if parallel_state.get_virtual_pipeline_model_parallel_world_size() is not None:
            if "state_dict" in checkpoint:
                checkpoint_state_dict = checkpoint["state_dict"][f"model_{index}"]
            else:
                checkpoint_state_dict = checkpoint[f"model_{index}"]
        else:
            if "state_dict" in checkpoint:
                checkpoint_state_dict = checkpoint["state_dict"]
            else:
                checkpoint_state_dict = checkpoint

        n_nesting = 0
        mcore_model = megatron_parallel.module
        while hasattr(mcore_model, "module"):
            mcore_model = mcore_model.module
            n_nesting += 1

        _state_dict = {}
        for key, value in checkpoint_state_dict.items():
            # Count the number of "module." at the start of the key
            count, _key = 0, key
            while _key.startswith("module."):
                _key = _key[len("module.") :]
                count += 1

            # Adjust the number of "module." prefixes
            if count < n_nesting:
                to_add = "module." * (n_nesting - count)
                _state_dict[f"{to_add}{key}"] = value
            elif count > n_nesting:
                to_remove = "module." * (count - n_nesting)
                _state_dict[key[len(to_remove) :]] = value
            else:
                _state_dict[key] = value

        if have_custom_fsdp and hasattr(module, "module") and isinstance(module.module, FullyShardedDataParallel):
            module.module.load_state_dict(_state_dict, strict=strict)
        elif have_megatron_fsdp and hasattr(module, "module") and isinstance(module.module, FullyShardedDataParallel):
            module.module.load_state_dict(_state_dict, strict=strict)
            continue

        try:
            module.load_state_dict(_state_dict, strict=strict)
        except RuntimeError as e:
            missing_keys, expected_keys = module.load_state_dict(checkpoint_state_dict, strict=False)
            if all(s.endswith('_extra_state') for s in missing_keys):
                logging.warning(
                    f'Loding checkpoint created with Transformer Engine version lower than 1.13. '
                    f'Missing layers {missing_keys} will be ignored.'
                )
            else:
                raise e


def _sync_from_last_pipeline_stage(value: torch.Tensor, broadcast: bool = False):
    """
    When pipeline parallelism is enabled,
    casts a tensor defined on the last pipeline stage to other ranks.

        Args:
            value (torch.Tensor): A tensor to be casted from the final pipeline stage of
                a pipeline parallelism group (e.g. loss).
                Note that this tensor should already be defined on the target rank(s) to fill with received data.
            broadcast (bool): When True, broadcasts value from the final pipeline stage rank to all ranks in its group.
                When False, only rank zero receives value from the final pipeline stage rank in its group.
                This mode exists to avoid slow one-to-many communication when not necessary. Defaults to False.
    """
    from megatron.core import parallel_state

    if parallel_state.get_pipeline_model_parallel_world_size() > 1:
        src_rank = parallel_state.get_pipeline_model_parallel_last_rank()

        if not broadcast:
            pp_ranks = torch.distributed.get_process_group_ranks(parallel_state.get_pipeline_model_parallel_group())
            if torch.distributed.get_rank() == src_rank and 0 in pp_ranks:
                torch.distributed.send(value, 0)
            elif torch.distributed.get_rank() == 0:
                torch.distributed.recv(value, src_rank)
        else:
            torch.distributed.broadcast(
                value,
                src_rank,
                group=parallel_state.get_pipeline_model_parallel_group(),
            )


def setup_megatron_optimizer(
    model,
    config,
    no_weight_decay_cond: Optional[Callable] = None,
    scale_lr_cond: Optional[Callable] = None,
    lr_mult: float = 1.0,
):
    """ """
    from megatron.core.optimizer import OptimizerConfig, get_megatron_optimizer

    from nemo.core.optim import McoreDistributedOptimizer
    from nemo.utils import AppState

    app_state = AppState()

    assert isinstance(config, OptimizerConfig), f"Expected OptimizerConfig, got {type(config)}"

    class McoreOpt(McoreDistributedOptimizer):
        """ """

        def sharded_state_dict(
            self,
            model_sharded_state_dict,
            optimizer_state_dict=None,
            is_loading=False,
            sharding_type='fully_sharded_model_space',
            metadata=None,
        ):
            mcore_optimizer_sig = inspect.signature(self.mcore_optimizer.sharded_state_dict).parameters
            distrib_optim_kwargs = {}
            if "metadata" in mcore_optimizer_sig or "kwargs" in mcore_optimizer_sig:
                distrib_optim_kwargs["metadata"] = metadata
            elif "sharding_type" in mcore_optimizer_sig:
                distrib_optim_kwargs["sharding_type"] = sharding_type
            state_dict = self.mcore_optimizer.sharded_state_dict(
                model_sharded_state_dict, is_loading=is_loading, **distrib_optim_kwargs
            )
            return state_dict

    # megatron optimizer expects McoreDDP
    ddp_modules = [m.module for m in model]
    mcore_opt = get_megatron_optimizer(
        config,
        ddp_modules,
        no_weight_decay_cond=no_weight_decay_cond,
        scale_lr_cond=scale_lr_cond,
        lr_mult=lr_mult,
        use_gloo_process_groups=app_state.use_gloo_process_groups,
    )
    # Pytorch does not have the concept of an `lr_mult` or a `wd_mult` but these are added to param
    # groups in megatron to control which sub-modules have different learning rates or weight
    # decays. Apply the multipliers here to each param_group's lr and wd, and to reduce confusion
    # change the name of these variables. We need this because nemo does not use the custom
    # megatron scheduler, and the megatron scheduler is what makes use of these mult parameters:
    # https://github.com/NVIDIA/Megatron-LM/blob/044e2ad5/megatron/core/optimizer_param_scheduler.py#L192-L193
    for pg in mcore_opt.param_groups:
        if 'pre_lr_mult' in pg or 'pre_mult_wd' in pg:
            # User has already applied custom lr and wd multipliers, don't apply `lr_mult` and
            # `wd_mult` again. This case may be encountered when resuming training.
            continue
        pg['pre_mult_lr'] = pg["lr"]
        pg['pre_mult_wd'] = pg['weight_decay']
        new_lr = pg["lr"] * pg.get('lr_mult', 1.0)
        new_wd = pg["weight_decay"] * pg.get("wd_mult", 1.0)
        pg['lr'] = new_lr
        pg['weight_decay'] = new_wd
        # In case a future implementation makes use of `lr_mult` and `wd_mult` directly in the
        #  scheduler, but accidentally also uses this function, remove `lr_mult` and `wd_mult` from
        #  the param groups so that the default value of 1.0 gets applied.
        if 'lr_mult' in pg:
            pg['pre_lr_mult'] = pg['lr_mult']
            del pg['lr_mult']  # remove so downstream methods do not apply again.
        if 'wd_mult' in pg:
            pg['pre_wd_mult'] = pg['wd_mult']
            del pg['wd_mult']  # remove so downstream methods do not apply again

    if getattr(model.ddp_config, "overlap_param_gather", False) and getattr(
        model.ddp_config, "align_param_gather", False
    ):
        param_sync_func = [model_chunk.start_param_sync for model_chunk in model]
        param_sync_func = param_sync_func[0] if len(model) == 1 else param_sync_func
        for module in model:
            module.config.param_sync_func = param_sync_func

    return McoreOpt(mcore_opt)



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from typing import List

import numpy as np
import pytest
import torch

from nemo.collections.asr.parts.k2.rnnt_logprobs import rnnt_logprobs_torch
from nemo.collections.asr.parts.numba.rnnt_loss.rnnt_numpy import RNNTLoss as RNNTLoss_Numpy
from nemo.core.utils.optional_libs import K2_AVAILABLE, TRITON_AVAILABLE

if K2_AVAILABLE:
    import k2

    from nemo.collections.asr.parts.k2.graph_transducer import GraphRnntLoss

if TRITON_AVAILABLE:
    from nemo.collections.asr.parts.k2.rnnt_logprobs_triton import rnnt_logprobs_triton


EPS_SM_INPUT = 1e-6
EPS_L_INPUT = 1e-4

DEVICES = ['cpu']

if K2_AVAILABLE and torch.cuda.is_available() and k2.with_cuda:
    DEVICES.append('cuda')


@pytest.mark.skipif(not K2_AVAILABLE, reason="k2 is not installed, skipping Graph-RNNT tests.")
class TestGraphRnnt:
    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("blank_first", [True, False])
    @pytest.mark.parametrize("num_frames", [1, 3, 6])
    @pytest.mark.parametrize("vocab_size", [3])
    def test_temporal_schema(self, device, blank_first, num_frames, vocab_size):
        blank_id = 0 if blank_first else vocab_size - 1
        loss = GraphRnntLoss(blank=blank_id)
        temporal_schema = loss.get_temporal_schema(
            num_frames=num_frames, vocab_size=vocab_size, device=torch.device(device)
        )

        etalon_schema_fst: List[List[int]] = []
        for time_i in range(num_frames):
            for label_i in range(vocab_size):
                if label_i == blank_id:
                    # transition to the next state
                    etalon_schema_fst.append([time_i, time_i + 1, label_i, time_i, 0])
                else:
                    # self-loop
                    etalon_schema_fst.append([time_i, time_i, label_i, time_i, 0])
        etalon_schema_fst.append([num_frames, num_frames + 1, -1, -1, 0])  # transition to final state
        etalon_schema_fst.append([num_frames + 1])  # final state
        etalon_schema_fst = sorted(etalon_schema_fst)  # required for k2.Fsa.from_str
        etalon_schema_fst_str = "\n".join([" ".join(map(str, line)) for line in etalon_schema_fst])
        etalon_temporal_schema = k2.Fsa.from_str(etalon_schema_fst_str, num_aux_labels=1)

        assert temporal_schema.num_arcs == etalon_temporal_schema.num_arcs
        assert temporal_schema.shape == etalon_temporal_schema.shape  # (num_states, None)
        assert k2.is_rand_equivalent(
            temporal_schema, etalon_temporal_schema, log_semiring=True, treat_epsilons_specially=False
        ), "Temporal schema mismatch"
        assert k2.is_rand_equivalent(
            temporal_schema.invert(),
            etalon_temporal_schema.invert(),
            log_semiring=True,
            treat_epsilons_specially=False,
        ), "Temporal schema output labels mismatch"

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("blank_first", [True, False])
    def test_unit_schema(self, device, blank_first):
        vocab_size = 3
        blank_id = 0 if blank_first else vocab_size - 1
        if blank_first:
            labels = [1, 1, 2, 1]
        else:
            labels = [1, 1, 0, 1]
        loss = GraphRnntLoss(blank=blank_id)
        unit_schema = loss.get_unit_schema(
            units_tensor=torch.tensor(labels, device=torch.device(device)), vocab_size=vocab_size
        )

        etalon_schema_fst: List[List[int]] = []
        for label_i, label in enumerate(labels):
            etalon_schema_fst.append([label_i, label_i + 1, label, label, label_i, 0])  # forward: label
            etalon_schema_fst.append([label_i, label_i, blank_id, blank_id, label_i, 0])  # self-loop: blank
        etalon_schema_fst.append([len(labels), len(labels), blank_id, blank_id, len(labels), 0])
        etalon_schema_fst.append([len(labels), len(labels) + 1, -1, -1, -1, 0])  # transition to final state
        etalon_schema_fst.append([len(labels) + 1])  # final state
        etalon_schema_fst = sorted(etalon_schema_fst)  # required for k2.Fsa.from_str
        etalon_schema_fst_str = "\n".join([" ".join(map(str, line)) for line in etalon_schema_fst])
        etalon_unit_schema = k2.Fsa.from_str(etalon_schema_fst_str, aux_label_names=["aux_labels", "unit_positions"])

        assert unit_schema.num_arcs == etalon_unit_schema.num_arcs
        assert unit_schema.shape == etalon_unit_schema.shape  # (num_states, None)
        assert k2.is_rand_equivalent(
            unit_schema, etalon_unit_schema, log_semiring=True, treat_epsilons_specially=False
        ), "Unit schema input labels mismatch"
        assert k2.is_rand_equivalent(
            unit_schema.invert(), etalon_unit_schema.invert(), log_semiring=True, treat_epsilons_specially=False
        ), "Unit schema output labels mismatch"

        # swap aux_labels and unit positions to test unit_positions
        unit_schema.aux_labels, unit_schema.unit_positions = unit_schema.unit_positions, unit_schema.aux_labels
        etalon_unit_schema.aux_labels, etalon_unit_schema.unit_positions = (
            etalon_unit_schema.unit_positions,
            etalon_unit_schema.aux_labels,
        )
        assert k2.is_rand_equivalent(
            unit_schema.invert(), etalon_unit_schema.invert(), log_semiring=True, treat_epsilons_specially=False
        ), "Unit schema unit positions mismatch"

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("blank_first", [True, False])
    def test_grid_schema(self, device, blank_first):
        vocab_size = 3
        blank_id = 0 if blank_first else vocab_size - 1
        if blank_first:
            labels = [1, 1, 2, 1]
        else:
            labels = [1, 1, 0, 1]
        text_length = len(labels)
        num_frames = 5
        loss = GraphRnntLoss(blank=blank_id)
        grid_schema = loss.get_grid(
            units_tensor=torch.tensor(labels, device=torch.device(device)),
            num_frames=num_frames,
            vocab_size=vocab_size,
        )

        etalon_schema_fst: List[List[int]] = []
        for frame_i in range(num_frames):
            for label_i in range(text_length + 1):
                state = frame_i * (text_length + 1) + label_i
                if label_i < text_length:
                    next_state_label = state + 1
                    # next unit
                    etalon_schema_fst.append([state, next_state_label, labels[label_i], frame_i, label_i, 0])
                if frame_i < num_frames - 1:
                    next_state_frame = (frame_i + 1) * (text_length + 1) + label_i
                    # next time frame (blank)
                    etalon_schema_fst.append([state, next_state_frame, blank_id, frame_i, label_i, 0])

        last_grid_state = num_frames * (text_length + 1) - 1
        etalon_schema_fst.append([last_grid_state, last_grid_state + 1, blank_id, num_frames - 1, text_length, 0])
        etalon_schema_fst.append(
            [last_grid_state + 1, last_grid_state + 2, -1, -1, -1, 0]
        )  # transition to final state
        etalon_schema_fst.append([last_grid_state + 2])  # final state
        etalon_schema_fst = sorted(etalon_schema_fst)  # required for k2.Fsa.from_str
        etalon_schema_fst_str = "\n".join([" ".join(map(str, line)) for line in etalon_schema_fst])
        etalon_grid_schema = k2.Fsa.from_str(etalon_schema_fst_str, aux_label_names=["aux_labels", "unit_positions"])

        assert grid_schema.num_arcs == etalon_grid_schema.num_arcs
        assert grid_schema.shape == etalon_grid_schema.shape  # (num_states, None)
        assert k2.is_rand_equivalent(
            grid_schema, etalon_grid_schema, log_semiring=True, treat_epsilons_specially=False
        ), "Grid schema input labels mismatch"
        assert k2.is_rand_equivalent(
            grid_schema.invert(), etalon_grid_schema.invert(), log_semiring=True, treat_epsilons_specially=False
        ), "Grid schema output labels mismatch"

        # swap aux_labels and unit positions to test unit_positions
        grid_schema.aux_labels, grid_schema.unit_positions = grid_schema.unit_positions, grid_schema.aux_labels
        etalon_grid_schema.aux_labels, etalon_grid_schema.unit_positions = (
            etalon_grid_schema.unit_positions,
            etalon_grid_schema.aux_labels,
        )
        assert k2.is_rand_equivalent(
            grid_schema.invert(), etalon_grid_schema.invert(), log_semiring=True, treat_epsilons_specially=False
        ), "Grid schema unit positions mismatch"

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("connect_composed", [True, False])
    @pytest.mark.parametrize("blank_first", [True, False])
    def test_small_compose_transducer(
        self, device, connect_composed, blank_first, rnnt_test_helper, rnn_loss_sample_data
    ):
        if blank_first:
            sample_data = rnn_loss_sample_data.get_sample_small()
        else:
            sample_data = rnn_loss_sample_data.get_sample_small_blank_last()
        graph_rnnt = GraphRnntLoss(
            blank=sample_data.blank_id, connect_composed=connect_composed, use_grid_implementation=False
        )
        graph_cost, graph_grads = rnnt_test_helper.wrap_and_call(
            graph_rnnt, sample_data.logits, sample_data.targets, device
        )
        assert np.allclose(graph_cost, sample_data.expected_cost.numpy(), rtol=EPS_SM_INPUT), "costs mismatch."
        assert np.allclose(graph_grads, sample_data.expected_grads.numpy(), atol=1e-6), "gradient mismatch."

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    def test_small_grid_transducer(self, device, rnnt_test_helper, rnn_loss_sample_data):
        sample_data = rnn_loss_sample_data.get_sample_small()
        graph_rnnt = GraphRnntLoss(blank=0, use_grid_implementation=True)
        graph_cost, graph_grads = rnnt_test_helper.wrap_and_call(
            graph_rnnt, sample_data.logits, sample_data.targets, device
        )
        assert np.allclose(graph_cost, sample_data.expected_cost.numpy(), rtol=EPS_SM_INPUT), "costs mismatch."
        assert np.allclose(graph_grads, sample_data.expected_grads.numpy(), atol=1e-6), "gradient mismatch."

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("use_triton", [True, False])
    def test_medium_grid_transducer(self, device, use_triton: bool, rnnt_test_helper, rnn_loss_sample_data):
        if use_triton and device == "cpu":
            pytest.skip("Triton does not support CPU yet")
        sample_data = rnn_loss_sample_data.get_sample_medium()
        graph_rnnt = GraphRnntLoss(blank=0, use_grid_implementation=True, use_triton=use_triton)
        graph_cost, graph_grads = rnnt_test_helper.wrap_and_call(
            graph_rnnt, sample_data.logits, sample_data.targets, device
        )
        assert np.allclose(graph_cost, sample_data.expected_cost.numpy(), rtol=EPS_SM_INPUT), "costs mismatch."
        assert np.allclose(graph_grads, sample_data.expected_grads.numpy(), atol=1e-6), "gradient mismatch."

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("use_triton", [True, False])
    def test_medium_random_var_size(self, device, use_triton: bool, rnnt_test_helper, rnn_loss_sample_data):
        if use_triton and device == "cpu":
            pytest.skip("Triton does not support CPU yet")
        sample_data = rnn_loss_sample_data.get_sample_medium_random_var_size(blank_first=True)
        graph_rnnt = GraphRnntLoss(blank=0, use_grid_implementation=True, use_triton=use_triton)
        graph_cost, graph_grads = rnnt_test_helper.wrap_and_call(
            graph_rnnt,
            sample_data.logits.detach(),
            sample_data.targets,
            device,
            input_lengths=sample_data.input_lengths,
            target_lengths=sample_data.target_lengths,
        )
        etalon_rnnt = RNNTLoss_Numpy(blank=0)
        etalon_cost, etalon_grads = rnnt_test_helper.wrap_and_call(
            etalon_rnnt,
            sample_data.logits.detach(),
            sample_data.targets,
            device,
            input_lengths=sample_data.input_lengths,
            target_lengths=sample_data.target_lengths,
        )
        assert np.allclose(graph_cost.sum(), etalon_cost, rtol=EPS_SM_INPUT), "costs mismatch."
        assert np.allclose(graph_grads, etalon_grads, atol=1e-4), "gradient mismatch."

    @pytest.mark.unit
    @pytest.mark.parametrize("device", DEVICES)
    @pytest.mark.parametrize("blank_first", [True, False])
    def test_small_random_grid_compose_equivalent(self, device: torch.device, blank_first: bool, rnn_loss_sample_data):
        sample_data = rnn_loss_sample_data.get_sample_small_random(blank_first, device=device)
        criterion = GraphRnntLoss(blank=sample_data.blank_id, connect_composed=True, use_grid_implementation=False)
        text_tensor = sample_data.targets[0]
        num_frames = sample_data.logits.shape[1]
        graph_grid = criterion.get_grid(text_tensor, num_frames, sample_data.vocab_size)
        graph_composed = criterion.get_composed_lattice(text_tensor, num_frames, sample_data.vocab_size)
        assert k2.is_rand_equivalent(
            graph_grid, graph_composed, log_semiring=True, treat_epsilons_specially=False
        ), "Grid and composed graphs are not equivalent."


@pytest.mark.skipif(not TRITON_AVAILABLE, reason="Triton is not installed, skipping RNNT Log Probs tests")
@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is unavailable")
class TestRnntLogProbs:
    @pytest.mark.parametrize(
        "batch_size,num_frames,num_text_units,vocab_size",
        [
            (1, 4, 2, 4),
            (2, 3, 2, 5),
            (2, 16, 31, 17),
            (16, 129, 65, 2048),
        ],
    )
    @pytest.mark.parametrize(
        "float_dtype",
        [torch.float32] + ([torch.bfloat16] if torch.cuda.is_available() and torch.cuda.is_bf16_supported() else []),
    )
    def test_rnnt_logprobs_random(
        self, batch_size: int, num_frames: int, num_text_units: int, vocab_size: int, float_dtype: torch.dtype
    ):
        """
        Test Triton-based implementation using etalon Torch-based implementation for RNN-T log-probs.
        """
        device = torch.device("cuda")
        torch.manual_seed(777)

        targets = torch.tensor(
            [[random.randrange(0, vocab_size - 1) for i in range(num_text_units)] for j in range(batch_size)],
            device=device,
            dtype=torch.long,
        )

        logits = torch.rand(
            [batch_size, num_frames, num_text_units + 1, vocab_size + 1],
            dtype=float_dtype,
            device=device,
            requires_grad=True,
        )

        # Triton-based implementation works in float32 precision for accuracy purposes, should compare with float32
        target_scores_etalon, blank_scores_etalon = rnnt_logprobs_torch(
            logits=logits.to(torch.float32), targets=targets, blank_id=vocab_size
        )
        logits2 = logits.clone().detach()
        logits2.requires_grad_(True)
        target_scores, blank_scores = rnnt_logprobs_triton(logits=logits2, targets=targets, blank_id=vocab_size)
        target_scores[..., -1:] = 0.0
        target_scores_etalon[..., -1:] = 0.0
        assert torch.allclose(blank_scores, blank_scores_etalon, atol=1e-5)
        assert torch.allclose(target_scores, target_scores_etalon, atol=1e-5)

        # test backward
        target_scales = torch.rand_like(target_scores, requires_grad=False)
        blank_scales = torch.rand_like(blank_scores, requires_grad=False)
        loss_etalon = (target_scales * target_scores_etalon + blank_scales * blank_scores_etalon).sum()
        loss = (target_scales * target_scores + blank_scales * blank_scores).sum()
        loss_etalon.backward()
        loss.backward()
        assert torch.allclose(logits.grad, logits2.grad, atol=1e-5)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field

from nemo.collections.asr.models.configs.asr_models_config import EncDecCTCConfig
from nemo.collections.asr.parts.k2.classes import GraphModuleConfig as BackendConfig
from nemo.core.config.modelPT import NemoConfig


@dataclass
class GraphModuleConfig:
    criterion_type: str = "ml"
    loss_type: str = "ctc"
    split_batch_size: int = 0
    dec_type: str = "topo"
    transcribe_training: bool = True
    backend_cfg: BackendConfig = field(default_factory=lambda: BackendConfig())


@dataclass
class EncDecK2SeqConfig(EncDecCTCConfig):
    graph_module_cfg: GraphModuleConfig = field(default_factory=lambda: GraphModuleConfig())


@dataclass
class EncDecK2SeqModelConfig(NemoConfig):
    model: EncDecK2SeqConfig = field(default_factory=lambda: EncDecK2SeqConfig())



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Callable, Optional

import lightning.pytorch as L
import torch

try:
    from megatron.core.extensions.transformer_engine import (
        TEColumnParallelLinear,
        TEDotProductAttention,
        TENorm,
        TERowParallelLinear,
    )
except ImportError:
    from nemo.utils import logging

    # These Defaults are needed to make sure the code compiles
    TEColumnParallelLinear = None
    TEDotProductAttention = object
    TENorm = None
    TERowParallelLinear = None

    logging.warning(
        "Failed to import Transformer Engine dependencies. "
        "`from megatron.core.extensions.transformer_engine import *`"
        "If using NeMo Run, this is expected. Otherwise, please verify the Transformer Engine installation."
    )

from megatron.core.parallel_state import (
    get_tensor_model_parallel_group,
    get_tensor_model_parallel_rank,
    get_tensor_model_parallel_world_size,
)
from megatron.core.tensor_parallel.layers import ColumnParallelLinear, RowParallelLinear
from megatron.core.transformer.attention import SelfAttention, SelfAttentionSubmodules
from megatron.core.transformer.dot_product_attention import DotProductAttention
from megatron.core.transformer.enums import AttnMaskType
from megatron.core.transformer.identity_op import IdentityOp
from megatron.core.transformer.mlp import MLP, MLPSubmodules
from megatron.core.transformer.spec_utils import ModuleSpec, build_module
from megatron.core.transformer.transformer_config import TransformerConfig
from megatron.core.transformer.transformer_layer import TransformerLayer, TransformerLayerSubmodules

from nemo.collections.vlm.vision.base import CLIPViTConfig
from nemo.lightning import io, teardown


class InternViTRMSNorm(torch.nn.Module):
    """Customized Version of RMSNorm"""

    def __init__(
        self,
        config,
        hidden_size: int,
        eps: float = 1e-6,
        sequence_parallel: bool = False,
        compute_var: bool = False,
    ):
        """Custom RMSNorm for InternViT.

        Args:
            config (TransformerConfig): Config.
            hidden_size (int): Input hidden size.
            eps (float): epsilon to use for the norm, default to 1e-6
            sequence_parallel (bool): Set to true if sequence parallelism is being used,
              this marks the weights as needing to be allreduced.
            compute_var (bool): Indicator to compute statistic manually.
        """
        super().__init__()
        self.config = config
        self.eps = eps
        self.weight = torch.nn.Parameter(torch.ones(hidden_size))
        self._compute_var = compute_var

        assert not sequence_parallel, "Sequence parallelism is not supported with InternViT."

        setattr(self.weight, 'sequence_parallel', sequence_parallel)

    def _norm(self, x, var):
        """RMSNorm"""
        if var is None:
            var = x.pow(2).mean(-1, keepdim=True)

        return x * torch.rsqrt(var + self.eps)

    def forward(self, x):
        """Run RMSNorm with an option to compute custom statistic."""
        var = None
        if self._compute_var:
            unpadded_hidden_size = self.config.hidden_size  # 3200
            max_dim = x.shape[-1]  # 128
            x = x.reshape(x.size(0), x.size(1), -1)
            total_heads = x.shape[-1] // max_dim
            if max_dim == 128:
                valid_heads = 25
            elif max_dim == 64:
                valid_heads = 16
            else:
                raise ValueError("Cannot infer number of heads.")
            var = self._gather_var(x.float().pow(2), max_dim, valid_heads, total_heads) / unpadded_hidden_size

        output = self._norm(x.float(), var).type_as(x)
        output = output * self.weight

        if self._compute_var:
            output = output.reshape(output.size(0), output.size(1), -1, max_dim)

        return output

    def _gather_var(self, input_, max_dim, valid_heads, total_heads):
        """Compute statistic across the non-dummy heads."""
        world_size = get_tensor_model_parallel_world_size()
        rank = get_tensor_model_parallel_rank()

        heads_per_rank = total_heads // world_size
        valid_heads_per_rank = []
        remaining_heads = valid_heads
        for _ in range(world_size):
            if remaining_heads >= heads_per_rank:
                valid_heads_per_rank.append(heads_per_rank)
            else:
                valid_heads_per_rank.append(remaining_heads)
            remaining_heads -= heads_per_rank

        # Size and dimension.
        last_dim = input_.dim() - 1

        valid_dim = max_dim * valid_heads_per_rank[rank]
        if valid_dim > 0:
            var = input_[..., :valid_dim].sum(-1, keepdim=True)
        else:
            var = input_.sum(-1, keepdim=True) * 0.0  # Zero-out the dummy heads.

        tensor_list = [torch.empty_like(var) for _ in range(world_size)]
        tensor_list[rank] = var
        torch.distributed.all_gather(tensor_list, var, group=get_tensor_model_parallel_group())

        output = torch.cat(tensor_list, dim=last_dim).contiguous()

        return output.sum(-1, keepdim=True)


def get_mlp_module_spec(use_te: bool = True) -> ModuleSpec:
    # pylint: disable=C0115,C0116
    # Dense MLP w/ or w/o TE modules.
    return ModuleSpec(
        module=MLP,
        submodules=MLPSubmodules(
            linear_fc1=TEColumnParallelLinear if use_te else ColumnParallelLinear,
            linear_fc2=TERowParallelLinear if use_te else RowParallelLinear,
        ),
    )


def _bias_dropout_add_func_internvit(ls, x_with_bias, residual, prob, training):
    """Handle InternViT's layer scaling."""
    x, bias = x_with_bias  # unpack
    residual = residual if residual.dtype == x.dtype else residual.to(x.dtype)
    if bias is not None:
        x = x + bias
        out = torch.nn.functional.dropout(x, p=prob, training=training)
        out = residual + out * ls
        return out
    else:
        out = torch.nn.functional.dropout(x, p=prob, training=training)
        out = residual + out * ls
        return out


def bias_dropout_add_unfused_internvit(ls, training):
    """Bias-dropout-add as in Megatron but with added LayerScaling handling."""

    def _bias_dropout_add(x_with_bias, residual, prob):
        #
        return _bias_dropout_add_func_internvit(ls, x_with_bias, residual, prob, training)

    return _bias_dropout_add


def get_bias_dropout_add_internvit(ls, training, fused):
    """Bias-dropout-add as in Megatron but with added LayerScaling handling."""
    assert not fused, "Fused bias-dropout-add not implemented for InternViT."
    return bias_dropout_add_unfused_internvit(ls, training)


class InternViTTransformerLayer(TransformerLayer):
    """Add InternViT specialties to our default TransformerLayer."""

    def __init__(self, *args, **kwargs):
        # pylint: disable=C0115,C0116
        super().__init__(*args, **kwargs)
        self.ls1 = torch.nn.Parameter(torch.ones(self.config.hidden_size))
        self.ls2 = torch.nn.Parameter(torch.ones(self.config.hidden_size))

        self.self_attn_bda = partial(self.self_attn_bda, self.ls1)
        self.mlp_bda = partial(self.mlp_bda, self.ls2)


class InternViTSelfAttention(SelfAttention):
    """Override a few things that are special in InternViT and not supported by the SelfAttention class."""

    def __init__(self, config: TransformerConfig, submodules: SelfAttentionSubmodules, *args, **kwargs):
        # pylint: disable=C0115,C0116
        super().__init__(config=config, submodules=submodules, *args, **kwargs)

        # Need to override linear_qkv, q_layernorm and k_layernorm.
        qkv_bias = self.config.add_qkv_bias

        self.linear_qkv = build_module(
            submodules.linear_qkv,
            self.config.hidden_size,
            self.query_projection_size + 2 * self.kv_projection_size,
            config=self.config,
            init_method=self.config.init_method,
            gather_output=False,
            bias=qkv_bias,
            skip_bias_add=False,
            is_expert=False,
            tp_comm_buffer_name='qkv',
        )

        qk_layernorm_hidden_size = (
            self.hidden_size_per_attention_head * self.num_attention_heads_per_partition
        )  # 512 for internvit
        self.q_layernorm = build_module(
            submodules.q_layernorm,
            hidden_size=qk_layernorm_hidden_size,
            config=self.config,
            eps=self.config.layernorm_epsilon,
            compute_var=True,
        )

        self.k_layernorm = build_module(
            submodules.k_layernorm,
            hidden_size=qk_layernorm_hidden_size,
            config=self.config,
            eps=self.config.layernorm_epsilon,
            compute_var=True,
        )


class InternViTTEDotProductAttention(TEDotProductAttention):
    """Adjusted Attention for InternViT"""

    def forward(self, *args, **kwargs):
        """Regular TEDotProductAttention + zero-out dummy attention heads."""
        out = super().forward(*args, **kwargs)

        # This makes sure the dummy attention heads are zeroed out.
        mask = torch.ones_like(out, dtype=out.dtype, device=out.device)
        rank = get_tensor_model_parallel_rank()
        max_dim = out.shape[-1]  # 128
        valid_ranks = 6

        if rank == valid_ranks:
            mask[..., max_dim:] *= 0.0
        elif rank > valid_ranks:
            mask *= 0.0
        out *= mask

        return out


def get_internvit_layer_spec(use_te, add_qk_norm=True, norm_type="RMSNorm") -> ModuleSpec:
    """Get InterViT's MCore layer spec"""
    NORM2FN = {
        'RMSNorm': InternViTRMSNorm,
        'LayerNorm': TENorm,
    }

    mlp = get_mlp_module_spec(use_te)  # no norm

    return ModuleSpec(
        module=InternViTTransformerLayer,
        submodules=TransformerLayerSubmodules(
            input_layernorm=NORM2FN[norm_type],
            self_attention=ModuleSpec(
                module=InternViTSelfAttention,
                params={"attn_mask_type": AttnMaskType.no_mask},
                submodules=SelfAttentionSubmodules(
                    linear_qkv=TEColumnParallelLinear if use_te else ColumnParallelLinear,
                    core_attention=TEDotProductAttention if use_te else DotProductAttention,
                    linear_proj=TERowParallelLinear if use_te else RowParallelLinear,
                    q_layernorm=NORM2FN[norm_type] if add_qk_norm else IdentityOp,
                    k_layernorm=NORM2FN[norm_type] if add_qk_norm else IdentityOp,
                ),
            ),
            self_attn_bda=get_bias_dropout_add_internvit,
            pre_mlp_layernorm=NORM2FN[norm_type],
            mlp=mlp,
            mlp_bda=get_bias_dropout_add_internvit,
        ),
    )


@dataclass
class InternViTConfig(CLIPViTConfig):
    """Intern ViT Base Config"""

    vision_model_type: str = "internvit"
    patch_dim: int = 14
    img_h: int = 448
    img_w: int = 448
    num_layers: int = 45
    num_attention_heads: int = 25
    num_query_groups: int = 25
    kv_channels: int = 128
    add_bias_linear: bool = True
    add_qkv_bias: bool = False
    hidden_size: int = 3200
    hidden_dropout: float = 0.0
    attention_dropout: float = 0.0
    ffn_hidden_size: int = 12800
    gated_linear_unit: bool = False
    activation_func: Callable = torch.nn.functional.gelu
    layernorm_zero_centered_gamma: bool = False
    apply_query_key_layer_scaling: bool = False
    bias_activation_fusion: bool = False
    bias_dropout_fusion: bool = False
    attention_softmax_in_fp32: bool = True
    normalization: str = 'RMSNorm'
    layernorm_epsilon: float = 1e-6
    apply_rope_fusion: bool = False
    transformer_layer_spec: ModuleSpec = field(default_factory=lambda: get_internvit_layer_spec(use_te=True))


@dataclass
class InternViT_6B_448px_Config(InternViTConfig):
    """Intern ViT 6B Config for >= v1.5"""

    vision_model_type: str = "internvit"


@dataclass
class InternViT_300M_448px_Config(InternViTConfig):
    """Intern ViT 300M Config for >= v1.5"""

    vision_model_type: str = "internvit"
    num_layers: int = 24
    num_attention_heads: int = 16
    num_query_groups: int = 16
    kv_channels: int = 64
    add_bias_linear: bool = True
    add_qkv_bias: bool = True
    hidden_size: int = 1024
    hidden_dropout: float = 0.0
    attention_dropout: float = 0.0
    ffn_hidden_size: int = 4096
    normalization: str = 'LayerNorm'
    transformer_layer_spec: ModuleSpec = field(
        default_factory=lambda: get_internvit_layer_spec(
            use_te=True,
            add_qk_norm=False,
            norm_type='LayerNorm',
        )
    )


class InternViTModel(L.LightningModule, io.IOMixin, io.ConnectorMixin):
    """InternViT NeMo Wrapper"""

    def __init__(self, config: Optional[InternViTConfig] = None):
        # pylint: disable=C0115,C0116
        super().__init__()
        self.config = config

    def configure_model(self) -> None:
        # pylint: disable=C0115,C0116
        if not hasattr(self, "module"):
            self.module = self.config.configure_model()


@io.model_importer(InternViTModel, "hf")
class HFInternViTImporter(io.ModelConnector["InternVisionModel", InternViTModel]):
    """HF InternViT Importer"""

    def init(self) -> InternViTModel:
        # pylint: disable=C0115,C0116
        return InternViTModel(self.config)

    def apply(self, output_path: Path) -> Path:
        # pylint: disable=C0115,C0116
        from transformers import AutoModel

        source = AutoModel.from_pretrained(str(self), trust_remote_code=True)
        target = self.init()
        trainer = self.nemo_setup(target)

        self.convert_state(source, target)
        print(f"Converted InternViT model to Nemo, saving to {output_path}")

        self.nemo_save(output_path, trainer)

        print(f"Converted InternViT model saved to {output_path}")

        teardown(trainer, target)
        del trainer, target

        return output_path

    def convert_state(self, source, target):
        # pylint: disable=C0115,C0116
        mapping = {
            # Embeddings
            "embeddings.class_embedding": "class_token",
            "embeddings.patch_embedding.weight": "conv1.weight",
            "embeddings.patch_embedding.bias": "conv1.bias",
            # Transformer Layers
            "encoder.layers.*.ls1": "decoder.layers.*.ls1",
            "encoder.layers.*.ls2": "decoder.layers.*.ls2",
            # Attention QKV
            "encoder.layers.*.attn.q_norm.weight": "decoder.layers.*.self_attention.q_layernorm.weight",
            "encoder.layers.*.attn.k_norm.weight": "decoder.layers.*.self_attention.k_layernorm.weight",
            "encoder.layers.*.attn.proj.weight": "decoder.layers.*.self_attention.linear_proj.weight",
            "encoder.layers.*.attn.proj.bias": "decoder.layers.*.self_attention.linear_proj.bias",
            # MLP
            "encoder.layers.*.mlp.fc1.weight": "decoder.layers.*.mlp.linear_fc1.weight",
            "encoder.layers.*.mlp.fc1.bias": "decoder.layers.*.mlp.linear_fc1.bias",
            "encoder.layers.*.mlp.fc2.weight": "decoder.layers.*.mlp.linear_fc2.weight",
            "encoder.layers.*.mlp.fc2.bias": "decoder.layers.*.mlp.linear_fc2.bias",
            # Layer Norm
            "encoder.layers.*.norm1.weight": "decoder.layers.*.input_layernorm.weight",
            "encoder.layers.*.norm1.bias": "decoder.layers.*.input_layernorm.bias",
            "encoder.layers.*.norm2.weight": "decoder.layers.*.pre_mlp_layernorm.weight",
            "encoder.layers.*.norm2.bias": "decoder.layers.*.pre_mlp_layernorm.bias",
        }

        return io.apply_transforms(
            source,
            target,
            mapping=mapping,
            transforms=[_import_position_embedding, _import_qkv, _import_qkv_bias],
        )

    @property
    def config(self) -> CLIPViTConfig:
        # pylint: disable=C0115,C0116
        from transformers import AutoConfig

        source = AutoConfig.from_pretrained(str(self), trust_remote_code=True)
        norm_type = getattr(source, "norm_type", "RMSNorm")
        if norm_type == "layer_norm":
            norm_type = "LayerNorm"
        output = InternViTConfig(
            patch_dim=source.patch_size,
            img_h=source.image_size,
            img_w=source.image_size,
            hidden_size=source.hidden_size,
            ffn_hidden_size=source.intermediate_size,
            layernorm_epsilon=source.layer_norm_eps,
            num_attention_heads=source.num_attention_heads,
            num_query_groups=source.num_attention_heads,
            kv_channels=source.hidden_size // source.num_attention_heads,
            add_qkv_bias=source.qkv_bias,
            num_layers=source.num_hidden_layers,
            normalization=norm_type,
            transformer_layer_spec=get_internvit_layer_spec(
                use_te=True,
                add_qk_norm=source.qk_normalization,
                norm_type=norm_type,
            ),
        )
        return output


@io.state_transform(
    source_key="embeddings.position_embedding",
    target_key="position_embeddings.weight",
)
def _import_position_embedding(ctx: io.TransformCTX, pos_emb):
    # pylint: disable=C0115,C0116
    return pos_emb.squeeze(0)


def import_qkv(q, k, v, head_num, num_query_groups, heads_per_group, hidden_size, head_size):
    # pylint: disable=C0115,C0116
    old_tensor_shape = q.size()
    new_q_tensor_shape = (head_num, head_size) + old_tensor_shape[1:]
    new_kv_tensor_shape = (num_query_groups, head_size) + old_tensor_shape[1:]

    q = q.view(*new_q_tensor_shape)
    k = k.view(*new_kv_tensor_shape)
    v = v.view(*new_kv_tensor_shape)

    qkv_weights_l = []
    for i in range(num_query_groups):
        qkv_weights_l.append(q[i * heads_per_group : (i + 1) * heads_per_group, :, :])
        qkv_weights_l.append(k[i : i + 1, :, :])
        qkv_weights_l.append(v[i : i + 1, :, :])
    qkv_weights = torch.cat(qkv_weights_l)
    assert qkv_weights.ndim == 3, qkv_weights.shape
    assert qkv_weights.shape[0] == (heads_per_group + 2) * num_query_groups, qkv_weights.shape
    assert qkv_weights.shape[1] == head_size, qkv_weights.shape
    assert qkv_weights.shape[2] == old_tensor_shape[1], qkv_weights.shape

    qkv_weights = qkv_weights.reshape([head_size * (head_num + 2 * num_query_groups), hidden_size])

    return qkv_weights


@io.state_transform(
    source_key="encoder.layers.*.attn.qkv.weight",
    target_key="decoder.layers.*.self_attention.linear_qkv.weight",
)
def _import_qkv(ctx: io.TransformCTX, qkv):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config
    q, k, v = qkv.chunk(3)
    return import_qkv(
        q,
        k,
        v,
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=megatron_config.hidden_size,
        head_size=megatron_config.kv_channels,
    )


@io.state_transform(
    source_key="encoder.layers.*.attn.qkv.bias",
    target_key="decoder.layers.*.self_attention.linear_qkv.bias",
)
def _import_qkv_bias(ctx: io.TransformCTX, qkv_bias):
    # pylint: disable=C0115,C0116
    megatron_config = ctx.target.config
    q_bias, k_bias, v_bias = qkv_bias.chunk(3)
    return import_qkv(
        q_bias.unsqueeze(-1),
        k_bias.unsqueeze(-1),
        v_bias.unsqueeze(-1),
        head_num=megatron_config.num_attention_heads,
        num_query_groups=megatron_config.num_query_groups,
        heads_per_group=megatron_config.num_attention_heads // megatron_config.num_query_groups,
        hidden_size=1,
        head_size=megatron_config.kv_channels,
    ).squeeze(-1)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import Optional

import torch

from nemo.utils.app_state import AppState

# pylint: disable=C0116


@functools.lru_cache(maxsize=None)
def _nvtx_enabled() -> bool:
    """Check if NVTX range profiling is enabled"""
    return AppState()._nvtx_ranges


# Messages associated with active NVTX ranges
_nvtx_range_messages: list[str] = []


def nvtx_range_push(msg: str) -> None:
    # Return immediately if NVTX range profiling is not enabled
    if not _nvtx_enabled():
        return

    # Push NVTX range to stack
    _nvtx_range_messages.append(msg)
    torch.cuda.nvtx.range_push(msg)


def nvtx_range_pop(msg: Optional[str] = None) -> None:
    # Return immediately if NVTX range profiling is not enabled
    if not _nvtx_enabled():
        return

    # Update list of NVTX range messages and check for consistency
    if not _nvtx_range_messages:
        raise RuntimeError("Attempted to pop NVTX range from empty stack")
    last_msg = _nvtx_range_messages.pop()
    if msg is not None and msg != last_msg:
        raise ValueError(
            f"Attempted to pop NVTX range from stack with msg={msg}, " f"but last range has msg={last_msg}"
        )

    # Pop NVTX range
    torch.cuda.nvtx.range_pop()


# pylint: enable=C0116



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Optional, Tuple, Union
import numpy as np
import tensorrt_llm
import torch
from tensorrt_llm._utils import mpi_comm, torch_to_numpy

# A global dicts to store exported weights.
# This is set to be a global variable to avoid extra code modification from tensorrt_llm.
weights_dict = {}


DECODER_MODEL_TYPE = {
    "gptj": 'GPTForCausalLM',
    "gptnext": 'GPTForCausalLM',
    "llama": 'LlamaForCausalLM',
    "gemma": 'GemmaForCausalLM',
    "falcon": 'FalconForCausalLM',
}

post_layernorm_keys = [
    "post_attention_layernorm.weight",
    "post_attention_layernorm.bias",
    "post_self_attn_layernorm.weight",
]
mlp_proj_bias_keys = ["mlp.linear_fc2.bias", "mlp.dense_4h_to_h.bias"]
attention_dense_bias_keys = ["attention.linear_proj.bias", "attention.dense.bias"]
input_layernorm_keys = ["input_layernorm.weight", "input_layernorm.bias"]
pre_layernorm_keys = ["pre_mlp_layernorm.weight", "pre_mlp_layernorm.bias"]
attention_dense_weight_keys = ["attention.linear_proj.weight", "attention.dense.weight"]
mlp_proj_weight_keys = ["mlp.linear_fc2.weight", "mlp.dense_4h_to_h.weight"]
mlp_fc_keys = ["mlp.dense_h_to_4h.weight", "mlp.dense_h_to_4h.bias", "mlp.linear_fc1.weight", "mlp.linear_fc1.bias"]
attention_qkv_bias_keys = ["attention.query_key_value.bias", "attention.linear_qkv.bias"]
attention_qkv_weight_keys = ["attention.query_key_value.weight", "attention.linear_qkv.weight"]
mlp_router_keys = ["mlp.router.weight"]
mlp_fc_expert_keys = ["experts.linear_fc1.weight"]
mlp_proj_experts_keys = ["experts.linear_fc2.weight"]
final_layernorm_keys = ["final_layernorm.weight", "final_layernorm.bias"]
mlp_dense_2_keys = ["mlp.dense_h_to_4h_2.weight", "mlp.dense_h_to_4h_2.bias"]
attention_not_mapped_keys = [
    "attention.query.weight",
    "attention.query.bias",
    "attention.key_value.weight",
    "attention.key_value.bias",
]

weight_scaling_suffix = '.weights_scaling_factor'
activation_scaling_suffix = '.activation_scaling_factor'


def save_val(val, dir, key, tp_num=None):
    suffix = "" if tp_num is None else f".{tp_num}.bin"
    global weights_dict

    # Transpose linear layer weights to the correct shape.
    if torch.is_tensor(val):
        val = val.detach().contiguous()
        if len(val.shape) >= 2:
            val = val.reshape(val.shape[0], -1)
            val = torch.transpose(val, 0, 1)
        if key not in weights_dict:
            weights_dict[f"{key}{suffix}"] = torch.empty(
                val.size(), dtype=val.dtype, layout=val.layout, device="cpu", pin_memory=True
            )
        weights_dict[f"{key}{suffix}"].copy_(val, non_blocking=True)
    else:
        if len(val.shape) >= 2:
            val = np.ascontiguousarray(np.transpose(val.reshape(val.shape[0], -1), [1, 0]))
        weights_dict[f"{key}{suffix}"] = val


def save_split(split_vals, dir, key, i, split_factor):
    for j, val in enumerate(split_vals):
        save_val(val, dir, key, i * split_factor + j)


def save_expert_split(split_vals, dir, key, i, split_factor):
    for j, val in enumerate(split_vals):
        tp_num = i * split_factor + j
        suffix = "" if tp_num is None else f".{tp_num}.bin"

        global weights_dict
        weights_dict[f"{key}{suffix}"] = val


def generate_int8(weights, act_range, is_qkv=False, multi_query_mode=False):
    """This function has two purposes:
    - compute quantized weights, scaled either per-tensor or per-column
    - compute scaling factors.

    Depending on the GEMM API (CUTLASS/CUBLAS) the required scaling factors differ.
    CUTLASS uses two sets of scaling factors. One for the activation X, one for the weight W.
    CUBLAS only has one (we can't do per-row scaling). So we must provide pre-multiplied scaling factor.

    Here is the list of what we need (T means per-tensor, C per-column):
    - scale_x_orig_quant puts fp activation into the quantized range (i.e. [-128, 127], for int8).
    Used before the GEMM. (T)
    - scale_y_quant_orig puts quantized activation into the fp range. Used if the GEMM outputs int8. (T)
    - scale_w_quant_orig puts weights from quant range to fp range (used with CUTLASS) (T, C)
    - scale_y_accum_quant puts the GEMM result (XW) from accumulation range (int32)
    to quant range (int8) (used for CUBLAS) (T, C)

    Note that we don't do anything special about row-parallel GEMM.
    Theoretically, we could have per-GPU scaling factors too,
    but then the model would change depending on the number of GPUs used.

    For QKV projection, the behavior is special. Even if we have a single matrix to perform QKV projection,
    we consider it
    as three different matrices: Q, K, and V. So per-tensor actually means one scaling factor for each Q, K and V.
    """
    # compute weight scaling factors for fp->int8 and int8->fp
    if is_qkv and not multi_query_mode:
        scale_w_orig_quant_t = 127.0 / act_range["w"].reshape(3, -1).max(dim=-1, keepdims=True)[0].cpu().numpy()
        scale_w_orig_quant_c = 127.0 / act_range["w"].reshape(3, -1).cpu().numpy()
    elif is_qkv and multi_query_mode:
        raise ValueError("Multi-query w/ int8 quant has not been supported yet")
    else:
        scale_w_orig_quant_t = 127.0 / act_range["w"].max().cpu().numpy()
        scale_w_orig_quant_c = 127.0 / act_range["w"].cpu().numpy()
    scale_w_quant_orig_t = 1.0 / scale_w_orig_quant_t
    scale_w_quant_orig_c = 1.0 / scale_w_orig_quant_c

    # compute the rest of needed scaling factors
    scale_x_orig_quant_t = np.array(127.0 / act_range["x"].max().item())
    scale_y_orig_quant_t = np.array(127.0 / act_range["y"].max().item())
    scale_y_quant_orig_t = np.array(act_range["y"].max().item() / 127.0)
    scale_y_accum_quant_t = scale_y_orig_quant_t / (scale_x_orig_quant_t * scale_w_orig_quant_t)
    scale_y_accum_quant_c = scale_y_orig_quant_t / (scale_x_orig_quant_t * scale_w_orig_quant_c)
    if is_qkv:
        scale_y_accum_quant_t = np.broadcast_to(scale_y_accum_quant_t, scale_w_orig_quant_c.shape)
        scale_w_quant_orig_t = np.broadcast_to(scale_w_quant_orig_t, scale_w_orig_quant_c.shape)

    def to_i8(x):
        return x.round().clip(-127, 127).astype(np.int8)

    return {
        "weight.int8": to_i8(weights * scale_w_orig_quant_t),
        "weight.int8.col": to_i8(weights * scale_w_orig_quant_c),
        "scale_x_orig_quant": scale_x_orig_quant_t.astype(np.float32),
        "scale_w_quant_orig": scale_w_quant_orig_t.astype(np.float32),
        "scale_w_quant_orig.col": scale_w_quant_orig_c.astype(np.float32),
        "scale_y_accum_quant": scale_y_accum_quant_t.astype(np.float32),
        "scale_y_accum_quant.col": scale_y_accum_quant_c.astype(np.float32),
        "scale_y_quant_orig": scale_y_quant_orig_t.astype(np.float32),
    }


def write_int8(vals, dir, base_key, split_dim, tp_rank, split_factor, kv_cache_only=False):
    if not kv_cache_only:
        save_split(
            np.split(vals["weight.int8"], split_factor, axis=split_dim),
            dir,
            f"{base_key}.weight.int8",
            tp_rank,
            split_factor,
        )
        save_split(
            np.split(vals["weight.int8.col"], split_factor, axis=split_dim),
            dir,
            f"{base_key}.weight.int8.col",
            tp_rank,
            split_factor,
        )

    saved_keys_once = ["scale_y_quant_orig"]
    if not kv_cache_only:
        saved_keys_once += ["scale_x_orig_quant", "scale_w_quant_orig", "scale_y_accum_quant"]
    # per-column scaling factors are loaded per-gpu for ColumnParallel GEMMs (QKV, FC1)
    if not kv_cache_only:
        if split_dim == -1:
            save_split(
                np.split(vals["scale_w_quant_orig.col"], split_factor, axis=split_dim),
                dir,
                f"{base_key}.scale_w_quant_orig.col",
                tp_rank,
                split_factor,
            )
            save_split(
                np.split(vals["scale_y_accum_quant.col"], split_factor, axis=split_dim),
                dir,
                f"{base_key}.scale_y_accum_quant.col",
                tp_rank,
                split_factor,
            )
        else:
            saved_keys_once += ["scale_w_quant_orig.col", "scale_y_accum_quant.col"]

    if tp_rank == 0:
        for save_key in saved_keys_once:
            save_val(vals[save_key], dir, f"{base_key}.{save_key}")


def get_suffix(key: str) -> str:
    return '.' + key.split('.')[-1]


def get_trt_llm_prefix(key: str) -> str:
    layer_num = key.split(".")[1]
    return f'transformer.layers.{layer_num}'


def any_word_in_key(key: str, words: List[str]) -> bool:
    return any([word in key for word in words])


def sequential_key_map(key: str, mapping: List[Tuple[List[str], str]]) -> Optional[str]:
    for keywords, mapped in mapping:
        if any_word_in_key(key, keywords):
            return mapped

    return None


def get_trt_llm_infix(key: str) -> Optional[str]:
    mapping = [
        (post_layernorm_keys, '.post_layernorm'),
        (mlp_proj_bias_keys, '.mlp.proj'),
        (attention_dense_bias_keys, '.attention.dense'),
        (input_layernorm_keys, '.input_layernorm'),
        (pre_layernorm_keys, '.post_layernorm'),
        (attention_dense_weight_keys, '.attention.dense'),
        (mlp_proj_weight_keys, '.mlp.proj'),
        (mlp_fc_keys, '.mlp.fc'),
        (attention_qkv_bias_keys + attention_qkv_weight_keys, '.attention.qkv'),
        (mlp_router_keys, '.mlp.router'),
        (mlp_fc_expert_keys, '.mlp.fc'),
        (mlp_proj_experts_keys, '.mlp.proj'),
    ]
    return sequential_key_map(key, mapping)


def get_trt_llm_keyname(key: str) -> str:
    if any_word_in_key(key, final_layernorm_keys):
        return key.replace("final_layernorm", "transformer.ln_f")

    if infix := get_trt_llm_infix(key):
        return get_trt_llm_prefix(key) + infix + get_suffix(key)

    return key


def is_scaling_factor(key: str) -> bool:
    return "scale_fwd" in key


def get_scaling_factor_keys(key: str) -> Tuple[Tuple[str, str], Tuple[str, str]]:
    # Reuses existing mapping of NeMo -> TRT LLM weights key via swapping suffixes
    corresponding_weight_key = '.'.join(key.split('.')[:-2]) + '.weight'
    corresponding_trt_llm_weight_key = get_trt_llm_keyname(corresponding_weight_key)
    base_key = '.'.join(corresponding_trt_llm_weight_key.split('.')[:-1])

    weight_scale = base_key + weight_scaling_suffix
    activation_scale = base_key + activation_scaling_suffix
    keys = (weight_scale, activation_scale)

    layer_prefix = get_trt_llm_prefix(key)
    mapped_key = layer_prefix + '.mlp.gate'
    gate_activation = mapped_key + activation_scaling_suffix
    gate_weight = mapped_key + weight_scaling_suffix
    gate_keys = (gate_activation, gate_weight)

    return keys, gate_keys


def save_scaling_factor(scaling_factors: dict, key: str, val: torch.Tensor, config: dict):
    if not is_scaling_factor(key):
        return scaling_factors

    activation_factor = 1 / val[0].view(1)
    weights_factor = 1 / val[1].view(1)

    (weights_key, activation_key), gate_keys = get_scaling_factor_keys(key)
    scaling_factors[activation_key] = activation_factor
    scaling_factors[weights_key] = weights_factor

    split_gated_activation = config.get("split_gated_activation", False)
    if split_gated_activation and any_word_in_key(key, ["mlp.dense_h_to_4h", "mlp.linear_fc1"]):
        (gate_activation_key, gate_weight_key) = gate_keys
        scaling_factors[gate_activation_key] = activation_factor
        scaling_factors[gate_weight_key] = weights_factor

    return scaling_factors


def cast_val_datatype(vals, trt_llm_key, storage_type, is_fp8_model, scaling_factors):
    if not is_fp8_model:
        return [val.to(storage_type) for val in vals]

    fp8_storage_type = torch.float8_e4m3fn
    quantized_keys = [
        k.split(weight_scaling_suffix)[0] for k in scaling_factors.keys() if k.endswith(weight_scaling_suffix)
    ]
    for k in quantized_keys:
        if k in trt_llm_key:
            storage_type = fp8_storage_type
            scale = scaling_factors[k + weight_scaling_suffix]
            vals = [val.to(torch.float32) / scale for val in vals]
            break

    return [val.to(storage_type) for val in vals]


def split_val_gate(vals: List[np.ndarray], convert_on_device: bool):
    if convert_on_device:
        return [[n] for n in torch.chunk(vals[0], 2, axis=-1)]

    splits = [np.split(val, 2, axis=-1) for val in vals]
    return list(zip(*splits))


# Note: in multi_query_mode, only query heads are split between multiple GPUs, while key/value head
# are not split as there is only one head per key/value.
@torch.no_grad()
def split_and_save_weight(
    tp_rank, saved_dir, split_factor, key, vals, storage_type, act_range, config, scaling_factors={}
):
    use_attention_nemo_shape = config.get("use_attention_nemo_shape", False)
    split_gated_activation = config.get("split_gated_activation", False)
    num_attention_heads = config.get("num_attention_heads", 0)
    tp_size = config.get("tp_size", 1)
    int8_outputs = config.get("int8_outputs", None)
    multi_query_mode = config.get("multi_query_mode", False)
    num_kv_heads = config.get("num_kv_heads", num_attention_heads)
    size_per_head = config.get("kv_channels", None)
    convert_on_device = config.get("convert_on_device", False)
    is_fp8_model = config.get("fp8_quantized", False)
    use_fp8_kv_cache = config.get("fp8_kvcache", False)
    save_int8 = int8_outputs == "all" or int8_outputs == "kv_cache_only"

    trt_llm_key = get_trt_llm_keyname(key)
    if not isinstance(vals, list):
        vals = [vals]

    if config.get("transpose_weights", False) and vals[0].ndim == 2:
        vals = [val.T for val in vals]
    if "layernorm.weight" in key and config.get("apply_layernorm_1p", False):
        vals = [val.float() + 1.0 for val in vals]

    vals = cast_val_datatype(vals, trt_llm_key, storage_type, is_fp8_model, scaling_factors)
    if convert_on_device:
        assert len(vals) == 1  # Should only convert a single device param per call
        assert torch.is_tensor(vals[0])
    elif torch.is_tensor(vals[0]):
        vals = [torch_to_numpy(val.cpu()) for val in vals]

    if any_word_in_key(
        key,
        input_layernorm_keys
        + pre_layernorm_keys
        + attention_dense_bias_keys
        + post_layernorm_keys
        + mlp_proj_bias_keys
        + final_layernorm_keys,
    ) and (tp_rank == 0 or convert_on_device):
        # shared weights, only need to convert the weights of rank 0
        save_val(vals[0], saved_dir, trt_llm_key)

    elif any_word_in_key(key, attention_dense_weight_keys + mlp_proj_weight_keys):
        if convert_on_device:
            save_val(vals[0], saved_dir, trt_llm_key)
        else:
            cat_dim = 0
            val = np.concatenate(vals, axis=cat_dim)
            split_vals = np.split(val, split_factor, axis=cat_dim)
            save_split(split_vals, saved_dir, trt_llm_key, tp_rank, split_factor)

        if act_range is not None and int8_outputs == "all":
            base_key = trt_llm_key.replace(".weight", "")
            vals_i8 = generate_int8(val, act_range, multi_query_mode=multi_query_mode)
            write_int8(vals_i8, saved_dir, base_key, cat_dim, tp_rank, split_factor)

    elif any_word_in_key(key, mlp_fc_keys):
        if split_gated_activation:
            vals, gates = split_val_gate(vals, convert_on_device)

        if convert_on_device:
            save_val(vals[0], saved_dir, trt_llm_key)
        else:
            cat_dim = -1
            val = np.concatenate(vals, axis=cat_dim)
            split_vals = np.split(val, split_factor, axis=cat_dim)
            save_split(split_vals, saved_dir, trt_llm_key, tp_rank, split_factor)

        if act_range is not None and int8_outputs == "all":
            base_key = trt_llm_key.replace(".weight", "")
            vals_i8 = generate_int8(val, act_range, multi_query_mode=multi_query_mode)
            write_int8(vals_i8, saved_dir, base_key, cat_dim, tp_rank, split_factor)

        if split_gated_activation:
            assert not save_int8
            layer_prefix = get_trt_llm_prefix(key)
            gate_key = layer_prefix + '.mlp.gate' + get_suffix(trt_llm_key)
            if convert_on_device:
                save_val(gates[0], saved_dir, gate_key)
            else:
                gate = np.concatenate(gates, axis=cat_dim)
                split_vals = np.split(gate, split_factor, axis=cat_dim)
                save_split(split_vals, saved_dir, gate_key, tp_rank, split_factor)

    elif any_word_in_key(key, mlp_dense_2_keys):
        if convert_on_device:
            save_val(vals[0], saved_dir, trt_llm_key)
        else:
            cat_dim = -1
            val = np.concatenate(vals, axis=cat_dim)
            split_vals = np.split(val, split_factor, axis=cat_dim)
            save_split(split_vals, saved_dir, trt_llm_key, tp_rank, split_factor)

        if act_range is not None and int8_outputs == "all":
            base_key = trt_llm_key.replace(".weight", "")
            vals_i8 = generate_int8(val, act_range, multi_query_mode=multi_query_mode)
            write_int8(vals_i8, saved_dir, base_key, cat_dim, tp_rank, split_factor)

    elif any_word_in_key(key, attention_qkv_bias_keys):
        qkv_hidden_dim = vals[0].shape[0]
        size_per_head = qkv_hidden_dim // (num_attention_heads + 2 * num_kv_heads)
        q_num = num_attention_heads // num_kv_heads

        # We first concat all sub weights per tp rank together.
        len_vals = len(vals)
        if convert_on_device:
            val = vals[0]
        else:
            val = np.concatenate(vals, axis=0)
        val = val.reshape(num_kv_heads * len_vals // tp_size, q_num + 2, size_per_head)

        # Split the QKV to separate variables.
        if convert_on_device:
            qkv = torch.split(val, [q_num, 1, 1], dim=1)
            split_vals = torch.concatenate([qkv[0].reshape(-1), qkv[1].reshape(-1), qkv[2].reshape(-1)], dim=1)
            save_val(split_vals, saved_dir, trt_llm_key)
        else:
            qkv = np.split(val, [q_num, q_num + 1], axis=1)
            q_split = np.split(qkv[0], split_factor, axis=0)
            k_split = np.split(qkv[1], split_factor, axis=0)
            v_split = np.split(qkv[2], split_factor, axis=0)

            # Concatenate Q, K, and V together
            split_vals = [
                np.concatenate([q_split[i].reshape(-1), k_split[i].reshape(-1), v_split[i].reshape(-1)], axis=0)
                for i in range(split_factor)
            ]
            save_split(split_vals, saved_dir, trt_llm_key, tp_rank, split_factor)

    elif any_word_in_key(key, attention_qkv_weight_keys):
        assert use_attention_nemo_shape, "Only support NEMO shape for QKV weights"
        hidden_dim = vals[0].shape[0]
        if size_per_head is None:
            size_per_head = hidden_dim // num_attention_heads
        q_num = num_attention_heads // num_kv_heads

        # When the merge factor exceeds 1, the 'vals' list will have multiple entries.
        # Depending on the format, 'vals' can look like either [QQQQ..KV, QQQQ..KV, ...](for GQA) or [QKV, QKV, ...](for MHA).
        # We first concat all sub weights per tp rank together.
        if convert_on_device:
            val = vals[0].reshape(hidden_dim, num_kv_heads // tp_size, q_num + 2, size_per_head)
            qkv = torch.split(val, [q_num, 1, 1], dim=2)
            split_vals = torch.concatenate(
                [qkv[0].reshape(hidden_dim, -1), qkv[1].reshape(hidden_dim, -1), qkv[2].reshape(hidden_dim, -1)], dim=1
            )
            save_val(split_vals, saved_dir, trt_llm_key)
        else:
            len_vals = len(vals)
            val = np.concatenate(vals, axis=1)
            val = val.reshape(hidden_dim, num_kv_heads * len_vals // tp_size, q_num + 2, size_per_head)

            # Split the QKV to separate variables.
            qkv = np.split(val, [q_num, q_num + 1], axis=2)

            query_groups_shape = qkv[0].shape
            if len(query_groups_shape) > 1:
                if (query_groups_shape[1] % split_factor) != 0:
                    raise Exception(
                        "Number of query groups of the models is {0}. Please select tensor parallelism size "
                        "that can split the number of query groups to equal number of query matrices in the "
                        "each GPU.".format(query_groups_shape[1])
                    )

            q_split = np.split(qkv[0], split_factor, axis=1)
            k_split = np.split(qkv[1], split_factor, axis=1)
            v_split = np.split(qkv[2], split_factor, axis=1)

            # Concatenate Q, K, and V together
            split_vals = [
                np.concatenate(
                    [
                        q_split[i].reshape(hidden_dim, -1),
                        k_split[i].reshape(hidden_dim, -1),
                        v_split[i].reshape(hidden_dim, -1),
                    ],
                    axis=1,
                )
                for i in range(split_factor)
            ]
            save_split(split_vals, saved_dir, trt_llm_key, tp_rank, split_factor)

        if save_int8:
            base_key = trt_llm_key.replace(".weight", "")
            vals_i8 = generate_int8(val, act_range, is_qkv=True, multi_query_mode=multi_query_mode)
            write_int8(
                vals_i8,
                saved_dir,
                base_key,
                cat_dim,
                tp_rank,
                split_factor,
                kv_cache_only=int8_outputs == "kv_cache_only",
            )

        if use_fp8_kv_cache:
            base_key = trt_llm_key.replace('.qkv.weight', '')
            scaling_factor = torch.FloatTensor([1.0])
            save_val(scaling_factor, dir, base_key + '.kv_cache_scaling_factor')

    elif any_word_in_key(key, attention_not_mapped_keys):
        pass

    elif any_word_in_key(key, mlp_router_keys):
        val = np.concatenate(vals, axis=1)
        save_val(val, saved_dir, trt_llm_key)

    elif any_word_in_key(key, mlp_fc_expert_keys):
        cat_dim = -1
        val = np.concatenate(vals, axis=cat_dim)
        w1, w3 = np.split(val, 2, axis=1)
        # w1 splits
        split_w1s = np.split(w1, split_factor, axis=1)
        # w3 splits
        split_w3s = np.split(w3, split_factor, axis=1)

        split_vals = [np.concatenate(item, axis=1) for item in zip(split_w3s, split_w1s)]
        save_expert_split(split_vals, saved_dir, trt_llm_key, tp_rank, split_factor)

    elif any_word_in_key(key, mlp_proj_experts_keys):
        cat_dim = -1
        val = np.concatenate(vals, axis=cat_dim)
        split_vals = np.split(val, split_factor, axis=cat_dim)
        save_expert_split(split_vals, saved_dir, trt_llm_key, tp_rank, split_factor)
    else:
        print(f"[WARNING] {key} not handled by converter")

    global weights_dict
    return weights_dict


def split(v: Union[np.ndarray, torch.Tensor], tp_size: int, idx: int, dim: int = 0):
    """Splits the np tensor v on dim and return the idx's slice."""
    if tp_size == 1:
        return v

    dim = dim if len(v.shape) != 1 else 0
    if torch.is_tensor(v):
        return torch.split(v, v.size(dim) // tp_size, dim=dim)[idx].contiguous()

    return np.ascontiguousarray(np.split(v, tp_size, axis=dim)[idx])


def init_model_parallel_from_nemo(reshard_model):
    from megatron.core import parallel_state

    pp_size = parallel_state.get_pipeline_model_parallel_world_size()
    tp_size = parallel_state.get_tensor_model_parallel_world_size()
    dp_size = parallel_state.get_data_parallel_world_size()
    tp_rank = parallel_state.get_tensor_model_parallel_rank()
    pp_rank = parallel_state.get_pipeline_model_parallel_rank()
    dp_rank = parallel_state.get_data_parallel_rank()

    if reshard_model and pp_size > 1:
        dp_size = dp_size * pp_size
        dp_rank = torch.distributed.get_rank() // tp_size
        pp_rank = 0
        pp_size = 1

    mp_rank = tp_size * pp_rank + tp_rank
    # Need to split cpp MPI World Comm because TensorRT-LLM NCCL plugins refer to the locally split comm.
    # High level call structure is: MpiComm::split -> MpiComm::setSession -> LOCAL_COMM_SESSION (used in allReducePlugin.cpp)
    tensorrt_llm.bindings.MpiComm.split(dp_rank, mp_rank)
    # Also split the python mpi communicator and set the global world one to the local split one
    new_comm = mpi_comm().Split(color=dp_rank, key=mp_rank)
    from mpi4py import MPI

    MPI.COMM_WORLD = new_comm

    return mp_rank, dp_rank, tp_size, pp_size, dp_size



# Copyright (c) 2022, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import pytest
import torch
import torch.nn as nn

from nemo.collections.asr.parts.submodules.batchnorm import (
    FusedBatchNorm1d,
    replace_bn_with_fused_bn,
    replace_bn_with_fused_bn_all,
)


class TestFusedBatchNorm1d:
    @pytest.mark.unit
    def test_constructor(self):
        num_features = 10
        fused_bn = FusedBatchNorm1d(num_features=num_features)
        assert fused_bn.weight.shape[0] == num_features
        assert fused_bn.bias.shape[0] == num_features
        # check initialization: weight is ones, bias is zeros (identity)
        assert torch.allclose(fused_bn.weight, torch.ones(num_features))
        assert torch.allclose(fused_bn.bias, torch.zeros(num_features))

    @pytest.mark.unit
    def test_from_batchnorm(self):
        num_features = 10

        # construct batchnorm
        bn = nn.BatchNorm1d(num_features=num_features)

        # update bn stats
        bn.train()
        batch_size = 4
        for _ in range(10):
            _ = bn(torch.rand(batch_size, num_features))

        # test eval mode is equivalent
        fused_bn = FusedBatchNorm1d.from_batchnorm(bn)
        bn.eval()

        sample_2d = torch.rand(batch_size, num_features)
        assert torch.allclose(bn(sample_2d), fused_bn(sample_2d))

        sample_3d = torch.rand(batch_size, num_features, 5)
        assert torch.allclose(bn(sample_3d), fused_bn(sample_3d))


class TestReplaceBNWithFusedBN:
    @pytest.mark.unit
    def test_replace_bn_with_fused_bn(self):
        model = nn.Sequential(
            OrderedDict(
                [
                    ("linear1", nn.Linear(1, 10)),
                    ("bn1", nn.BatchNorm1d(10)),
                    ("relu1", nn.ReLU()),
                    ("linear2", nn.Linear(10, 11)),
                    ("bn2", nn.BatchNorm1d(11)),
                    (
                        "submodule1",
                        nn.Sequential(OrderedDict([("linear3", nn.Linear(11, 12)), ("bn3", nn.BatchNorm1d(12))])),
                    ),
                ]
            )
        )
        replace_bn_with_fused_bn(model, "submodule1.bn3")
        assert isinstance(model.bn1, nn.BatchNorm1d)
        assert isinstance(model.bn2, nn.BatchNorm1d)
        assert isinstance(model.submodule1.bn3, FusedBatchNorm1d)

    @pytest.mark.unit
    def test_replace_bn_with_fused_bn_all(self):
        model = nn.Sequential(
            OrderedDict(
                [
                    ("linear1", nn.Linear(1, 10)),
                    ("bn1", nn.BatchNorm1d(10)),
                    ("relu1", nn.ReLU()),
                    ("linear2", nn.Linear(10, 11)),
                    ("bn2", nn.BatchNorm1d(11)),
                    (
                        "submodule1",
                        nn.Sequential(OrderedDict([("linear3", nn.Linear(11, 12)), ("bn3", nn.BatchNorm1d(12))])),
                    ),
                ]
            )
        )
        replace_bn_with_fused_bn_all(model)
        assert isinstance(model.bn1, FusedBatchNorm1d)
        assert isinstance(model.bn2, FusedBatchNorm1d)
        assert isinstance(model.submodule1.bn3, FusedBatchNorm1d)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import socket
from typing import Any, Callable, Dict, Optional

import torch
from lightning_fabric.plugins import CheckpointIO
from lightning_fabric.utilities.types import _PATH
from megatron.core.dist_checkpointing.tensor_aware_state_dict import MCoreTensorAwareStateDict
from nvidia_resiliency_ext.checkpointing.local.base_state_dict import TensorAwareStateDict
from nvidia_resiliency_ext.checkpointing.local.ckpt_managers.base_manager import BaseCheckpointManager
from nvidia_resiliency_ext.checkpointing.local.ckpt_managers.local_manager import LocalCheckpointManager
from nvidia_resiliency_ext.checkpointing.local.replication.strategies import LazyCliqueReplicationStrategy
from nvidia_resiliency_ext.fault_tolerance.dict_utils import dict_list_map_inplace
from nvidia_resiliency_ext.ptl_resiliency.local_checkpoint_callback import (
    HierarchicalCheckpointIO,
    LocalCheckpointCallback,
)

from nemo.lightning.pytorch.trainer import Trainer
from nemo.utils.callbacks.dist_ckpt_io import AsyncCompatibleCheckpointIO, AsyncFinalizableCheckpointIO

logger = logging.getLogger(__name__)


class MCoreHierarchicalCheckpointIO(HierarchicalCheckpointIO, AsyncCompatibleCheckpointIO):
    """HierarchicalCheckpointIO implementation compatible with MCore distributed checkpointing.

    Args:
        wrapped_checkpoint_io (CheckpointIO): previously used checkpoint_io (for global checkpoints).
        local_ckpt_manager (BaseCheckpointManager): local checkpoint manager used to store the local checkpoints
        get_global_ckpt_iteration_fn (Callable[[_PATH], int]): a function that retrieves the iteration
            of a global checkpoint that will be compared with local checkpoint iteration
            in order to decide which to resume from.
        async_save (bool, optional): enables asynchronous save. Passed down to the local checkpoint
            manager unless overriden with `local_ckpt_options` in `_save_local_checkpoint`.
            If True, MCoreHierarchicalCheckpointIO must be wrapped with `AsyncFinalizableCheckpointIO` wrapper
        local_ckpt_algo (str, optional): local checkpoint save algorithm. See MCoreTensorAwareStateDict for details.
            By default, uses a fully parallel save and load algorithm ('fully_parallel`).
        parallelization_group (ProcessGroup, optional): save/load parallelization group
        allow_cache (bool, optional): if True, subsequent checkpoint saves will reuse
            the cached parallelization metadata.
    """

    def __init__(
        self,
        wrapped_checkpoint_io: CheckpointIO,
        local_ckpt_manager: BaseCheckpointManager,
        get_global_ckpt_iteration_fn: Callable[[_PATH], int],
        async_save: bool = False,
        local_ckpt_algo: str = "fully_parallel",
        parallelization_group: Optional[torch.distributed.ProcessGroup] = None,
        allow_cache: bool = False,
    ):
        super().__init__(wrapped_checkpoint_io, local_ckpt_manager, get_global_ckpt_iteration_fn, async_save)
        self.local_ckpt_algo = local_ckpt_algo
        self.parallelization_group = parallelization_group
        self.cached_metadata = None
        self.allow_cache = allow_cache

    def to_tensor_aware_state_dict(self, checkpoint: Dict[str, Any]) -> TensorAwareStateDict:
        """Specialized implementation using MCoreTensorAwareStateDict.

        Wraps the state dict in MCoreTensorAwareStateDict and makes sure
        that "common" state dict doesn't have any CUDA tensors.
        """
        state_dict_for_save, _ = MCoreTensorAwareStateDict.from_state_dict(
            checkpoint,
            algo=self.local_ckpt_algo,
            parallelization_group=self.parallelization_group,
            cached_metadata=self.cached_metadata,
        )

        def to_cpu(x):
            if isinstance(x, torch.Tensor) and x.device.type != "cpu":
                logger.debug("Moving CUDA tensor to CPU")
                x = x.to("cpu", non_blocking=True)
            return x

        dict_list_map_inplace(to_cpu, state_dict_for_save.common)
        if self.allow_cache:
            self.cached_metadata = None
        return state_dict_for_save

    def from_tensor_aware_state_dict(
        self, tensor_aware_checkpoint: TensorAwareStateDict, sharded_state_dict=None, strict=None
    ):
        """Unwraps MCoreTensorAwareStateDict to a plain state dict."""
        assert isinstance(
            tensor_aware_checkpoint, MCoreTensorAwareStateDict
        ), f"Unexpected tensor aware state dict type: {type(tensor_aware_checkpoint)}"
        if strict is not None:
            logger.warning("MCoreTensorAwareStateDict does not yet support the 'strict' argument.")

        return tensor_aware_checkpoint.to_state_dict(
            sharded_state_dict,
            algo=self.local_ckpt_algo,
            parallelization_group=self.parallelization_group,
        )


def update_trainer_local_checkpoint_io(
    trainer: Trainer,
    local_checkpoint_base_dir: str,
    get_global_ckpt_iteration_fn: Callable[[_PATH], int],
    **kwargs,
) -> None:
    """Update the Trainer with the corresponding MCoreHierarchicalCheckpointIO if local checkpointing is used.

    Args:
        trainer (nl.Trainer): Trainer object to drive training loop.
        local_checkpoint_base_dir (str): Root directory under which to save local checkpoints.
        get_global_ckpt_iteration_fn (Callable): a function that retrieves the iteration of a global checkpoint
            that will be compared with local checkpoint iteration in order to decide which to resume from.
        **kwargs (dict): Additional kwargs passed to initialize MCoreHierarchicalCheckpointIO.

    Note:
        Async saving of local checkpoints is inferred based on what was configured on the strategy, if available.

    """
    callbacks = trainer.callbacks
    use_local_ckpt = any(isinstance(cb, LocalCheckpointCallback) for cb in callbacks)
    if not use_local_ckpt:
        return

    checkpoint_io = trainer.strategy.checkpoint_io
    # Infer async save setting for local checkpoint
    # based on whether async saving is configured for saving the global checkpoints
    async_save = getattr(trainer.strategy, "async_save", False)
    if async_save:
        # Access inner checkpoint IO
        assert isinstance(checkpoint_io, AsyncFinalizableCheckpointIO), type(checkpoint_io)
        checkpoint_io = checkpoint_io.checkpoint_io

    if trainer.num_nodes > 1:
        repl_strategy = LazyCliqueReplicationStrategy()
    else:
        # No replication for single node
        repl_strategy = None

    local_ckpt_manager = LocalCheckpointManager(
        os.path.join(local_checkpoint_base_dir, "local_ckpt", socket.gethostname()),
        repl_strategy=repl_strategy,
    )
    hierarchical_checkpointing_io = MCoreHierarchicalCheckpointIO(
        checkpoint_io,
        local_ckpt_manager,
        get_global_ckpt_iteration_fn,
        async_save=async_save,
        **kwargs,
    )

    if async_save:
        hierarchical_checkpointing_io = AsyncFinalizableCheckpointIO(hierarchical_checkpointing_io)

    trainer.strategy.checkpoint_io = hierarchical_checkpointing_io



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple
import torch
from tqdm import tqdm
from nemo.collections.asr.parts.utils.offline_clustering import (
    SpeakerClustering,
    get_scale_interpolated_embs,
    getCosAffinityMatrix,
    split_input_data,
)
from nemo.collections.asr.parts.utils.online_clustering import get_merge_quantity, run_reducer


class LongFormSpeakerClustering(torch.nn.Module):
    def __init__(self, cuda: bool = False):
        """
        Initializes a speaker clustering class tailored for long-form audio, leveraging methods from the `SpeakerClustering` class.
        The clustering algorithm for long-form content is executed via the `forward_infer` function (not shown here). Input embedding 
        vectors are divided into chunks, each of size `embeddings_per_chunk`. Within every chunk, the clustering algorithm aims 
        to identify `chunk_cluster_count` distinct clusters. The resulting clustering labels are then expanded to match the original 
        length of the input embeddings.
        
        NOTE: torch.jit.script currently does not support inherited methods with a `super()` call.

        Args:
            cuda (bool):
                Flag indicating whether CUDA is available for computation.
        """
        super().__init__()
        self.speaker_clustering = SpeakerClustering(cuda=cuda)
        self.embeddings_in_scales: List[torch.Tensor] = [torch.tensor([0])]
        self.timestamps_in_scales: List[torch.Tensor] = [torch.tensor([0])]
        self.cuda = cuda
        self.device = torch.device("cuda") if self.cuda else torch.device("cpu")

    def check_input(self, embeddings_per_chunk: int, chunk_cluster_count: int, max_num_speakers: int) -> None:
        """
        Checks the validity of the input parameters.
        
        Args:
            embeddings_per_chunk (int):
                The size of the windows in which the algorithm aims to identify `chunk_cluster_count` clusters.
            chunk_cluster_count (int):
                The target number of clusters to identify within each window.
            max_num_speakers (int):
                The maximum number of speakers to be detected in the audio.
        """
        if chunk_cluster_count is None or embeddings_per_chunk is None:
            raise ValueError(
                f"chunk_cluster_count ({chunk_cluster_count}) and embeddings_per_chunk ({embeddings_per_chunk}) should be set."
            )
        elif (
            all(v is not None for v in [chunk_cluster_count, embeddings_per_chunk])
            and chunk_cluster_count >= embeddings_per_chunk
        ):
            raise ValueError(
                f"chunk_cluster_count ({chunk_cluster_count}) should be smaller than embeddings_per_chunk ({embeddings_per_chunk})."
            )

        if chunk_cluster_count <= max_num_speakers:
            raise ValueError(
                f"chunk_cluster_count ({chunk_cluster_count}) should be larger than max_num_speakers ({max_num_speakers})."
            )

    def unpack_labels(
        self,
        Y_aggr: torch.Tensor,
        window_range_list: List[List[int]],
        absolute_merge_mapping: List[List[torch.Tensor]],
        org_len: int,
    ) -> torch.LongTensor:
        """
        Unpack the labels from the aggregated labels to the original labels.

        Args:
            Y_aggr (Tensor): 
                Aggregated label vector from the merged segments.
            window_range_list (List[List[int]]): 
                List of window ranges for each of the merged segments.
            absolute_merge_mapping (List[List[torch.Tensor]]): 
                List of absolute mappings for each of the merged segments. Each list element contains two tensors:
                    - The first tensor represents the absolute index of the bypassed segment (segments that remain unchanged).
                    - The second tensor represents the absolute index of the merged segment (segments that have had their indexes changed).
            org_len (int): 
                Original length of the labels. In most cases, this is a fairly large number (on the order of 10^5).

        Returns:
            Y_unpack (Tensor): 
                Unpacked labels derived from the aggregated labels.
        """
        Y_unpack = torch.zeros((org_len,)).long().to(Y_aggr.device)
        for (win_rng, abs_mapping) in zip(window_range_list, absolute_merge_mapping):
            inferred_merged_embs = Y_aggr[win_rng[0] : win_rng[1]]
            if len(abs_mapping[1]) > 0:
                Y_unpack[abs_mapping[1]] = inferred_merged_embs[-1].clone()  # Merged
                if len(abs_mapping[0]) > 0:
                    Y_unpack[abs_mapping[0]] = inferred_merged_embs[:-1].clone()  # Bypass
            else:
                if len(abs_mapping[0]) > 0:
                    Y_unpack[abs_mapping[0]] = inferred_merged_embs.clone()
        return Y_unpack

    def split_embs_to_windows(
        self, index: int, emb: torch.Tensor, embeddings_per_chunk: int,
    ) -> Tuple[torch.Tensor, int]:
        """
        Splits the embedding tensor into smaller window-sized tensors based on a given index.
        
        Args:
            index (int): The index of the desired window. This determines the starting point 
                         of the window using the formula:
                         start = embeddings_per_chunk * index
            emb (Tensor): The embedding tensor which needs to be split.
            embeddings_per_chunk (int):
                The size of the windows in which the algorithm aims to identify `chunk_cluster_count` clusters.

        Returns:
            emb_part (Tensor): 
                The window-sized tensor, which is a portion of the `emb`.
            offset_index (int): 
                The starting position of the window in the `emb` tensor.
        """
        if embeddings_per_chunk * (index + 1) > emb.shape[0]:
            emb_part = emb[-1 * embeddings_per_chunk :]
            offset_index = emb.shape[0] - embeddings_per_chunk
        else:
            emb_part = emb[embeddings_per_chunk * index : embeddings_per_chunk * (index + 1)]
            offset_index = embeddings_per_chunk * index
        return emb_part, offset_index

    def forward(self, param_dict: Dict[str, torch.Tensor]) -> torch.LongTensor:
        """
        A function wrapper designed for performing inference using an exported script format.

        Note:
            A dictionary is used to facilitate inference with the exported jit model in the Triton server. 
            This is done using an easy-to-understand naming convention.
            See https://github.com/triton-inference-server/server/blob/main/docs/user_guide/model_configuration.md#special-conventions-for-pytorch-backend

        Args:
            param_dict (dict):
                Dictionary containing the arguments for speaker clustering.
                See `forward_infer` function for the argument information.

        Returns:
            (LongTensor): Speaker labels for the segments in the given input embeddings.
        """
        embeddings_in_scales = param_dict['embeddings']
        timestamps_in_scales = param_dict['timestamps']
        multiscale_segment_counts = param_dict['multiscale_segment_counts']
        multiscale_weights = param_dict['multiscale_weights']
        oracle_num_speakers = int(param_dict['oracle_num_speakers'].item())
        max_num_speakers = int(param_dict['max_num_speakers'].item())
        enhanced_count_thres = int(param_dict['enhanced_count_thres'].item())
        sparse_search_volume = int(param_dict['sparse_search_volume'].item())
        max_rp_threshold = float(param_dict['max_rp_threshold'].item())
        fixed_thres = float(param_dict['fixed_thres'].item())
        return self.forward_infer(
            embeddings_in_scales=embeddings_in_scales,
            timestamps_in_scales=timestamps_in_scales,
            multiscale_segment_counts=multiscale_segment_counts,
            multiscale_weights=multiscale_weights,
            oracle_num_speakers=oracle_num_speakers,
            max_rp_threshold=max_rp_threshold,
            max_num_speakers=max_num_speakers,
            enhanced_count_thres=enhanced_count_thres,
            sparse_search_volume=sparse_search_volume,
            fixed_thres=fixed_thres,
        )

    def get_div_ceil_count(self, numer: int, denomin: int) -> int:
        """
        Calculates the ceiling of the division of two integers.
        
        Args:
            numer (int): Numerator, the number of segments or clusters, for example.
            denomin (int): Denominator, the number of speakers or clusters, for example.

        Returns:
            (int): The ceiling of the division of the two integers (number of chunks).
        """
        return int(torch.ceil(torch.tensor(numer / denomin)).item())

    def long_forward_infer(
        self,
        embeddings_in_scales: torch.Tensor,
        timestamps_in_scales: torch.Tensor,
        multiscale_segment_counts: torch.LongTensor,
        multiscale_weights: torch.Tensor,
        oracle_num_speakers: int,
        max_rp_threshold: float,
        max_num_speakers: int,
        sparse_search_volume: int,
        fixed_thres: float,
        chunk_cluster_count: int,
        embeddings_per_chunk: int,
    ) -> torch.LongTensor:
        """
        This is forward function for long-form speaker clustering.
        Please refer to `SpeakerClustering` class for the original argument information.
        
        In the `LongFormSpeakerClustering` process:
            Step-1: Input embeddings are divided into smaller windows of size `embeddings_per_chunk`.
            Step-2: Each window undergoes overclustering, resulting in `chunk_cluster_count` fine-grained clusters.
            Step-3: These fine-grained clusters are merged to form the aggregated clustering labels `Y_aggr`.
            Step-4: The `unpack_labels` function is then employed to map the aggregated labels `Y_aggr` back to the 
            original labels for all `org_len` input embeddings: `Y_unpack`.
        
        Args:
            embeddings_in_scales (Tensor):
                List containing concatenated Torch tensor embeddings across multiple scales.
                The length of the list is equal to the number of scales.
                Each tensor has dimensions of (Number of base segments) x (Embedding Dimension).
            timestamps_in_scales (Tensor):
                List containing concatenated Torch tensor timestamps across multiple scales.
                The length of the list is equal to the number of scales.
                Each tensor has dimensions of (Total number of segments across all scales) x 2.
                Example:
                    >>> timestamps_in_scales[0] = \
                        torch.Tensor([[0.4, 1.4], [0.9, 1.9], [1.4, 2.4], ... [121.2, 122.2]])
            multiscale_segment_counts (LongTensor):
                A Torch tensor containing the number of segments for each scale.
                The tensor has dimensions of (Number of scales).
                Example:
                    >>> multiscale_segment_counts = torch.LongTensor([31, 52, 84, 105, 120])
            multiscale_weights (Tensor):
                Multi-scale weights used when merging affinity scores.
                Example:
                    >>> multiscale_weights = torch.tensor([1.4, 1.3, 1.2, 1.1, 1.0])
            oracle_num_speakers (int):
                The number of speakers in a session as given by the reference transcript.
            max_num_speakers (int):
                The upper bound for the number of speakers in each session.
            max_rp_threshold (float):
                Limits the range of parameter search.
                The clustering performance can vary based on this range.
                The default value is 0.15.
            enhanced_count_thres (int):
                For shorter audio recordings, the clustering algorithm might not accumulate enough speaker profiles for each cluster.
                Thus, the function `getEnhancedSpeakerCount` uses anchor embeddings (dummy representations) to mitigate the effects of cluster sparsity.
                A value of 80 is recommended for `enhanced_count_thres`.
            sparse_search_volume (int):
                The number of p_values considered during NME analysis.
                The default is 30. Lower values speed up the NME-analysis but might lead to poorer parameter estimations. Values below 20 are not recommended.
            fixed_thres (float):
                If a `fixed_thres` value is provided, the NME-analysis process will be skipped.
                This value should be optimized on a development set for best results.
                By default, it is set to -1.0, and the function performs NME-analysis to estimate the threshold.
            kmeans_random_trials (int):
                The number of random trials for initializing k-means clustering. More trials can result in more stable clustering. The default is 1.
            chunk_cluster_count (int):
                The target number of clusters to identify within each chunk.
            embeddings_per_chunk (int):
                The size of the chunks in which the algorithm aims to identify `chunk_cluster_count` clusters.

        Returns:
            Y_unpack (LongTensor):
                Speaker labels for the segments in the provided input embeddings.
        """
        self.check_input(embeddings_per_chunk, chunk_cluster_count, max_num_speakers)

        self.embeddings_in_scales, self.timestamps_in_scales = split_input_data(
            embeddings_in_scales, timestamps_in_scales, multiscale_segment_counts
        )
        emb, _ = get_scale_interpolated_embs(
            multiscale_weights, self.embeddings_in_scales, self.timestamps_in_scales, self.device
        )
        offset_index: int = 0
        window_offset: int = 0
        total_emb: List[torch.Tensor] = []
        window_range_list: List[List[int]] = []
        absolute_merge_mapping: List[List[torch.Tensor]] = []
        total_window_count = self.get_div_ceil_count(numer=emb.shape[0], denomin=embeddings_per_chunk)

        if not torch.jit.is_scripting():
            pbar = tqdm(range(total_window_count), desc="Clustering Sub-Windows", leave=True, unit="window")
        else:
            pbar = range(total_window_count)

        for win_index in pbar:
            # Step-1: Split the embeddings into smaller chunks
            emb_part, offset_index = self.split_embs_to_windows(
                index=win_index, emb=emb, embeddings_per_chunk=embeddings_per_chunk
            )

            # Step-2: Perform overclustering on the chunks to identify `chunk_cluster_count` clusters
            if emb_part.shape[0] == 1:
                Y_part = torch.zeros((1,), dtype=torch.int64)
            else:
                mat = getCosAffinityMatrix(emb_part)
                overcluster_count = min(chunk_cluster_count, mat.shape[0])
                Y_part = self.speaker_clustering.forward_unit_infer(
                    mat=mat,
                    oracle_num_speakers=overcluster_count,
                    max_rp_threshold=max_rp_threshold,
                    max_num_speakers=chunk_cluster_count,
                    sparse_search_volume=sparse_search_volume,
                )

            # Step-3: Merge the clusters to form the aggregated clustering labels `Y_aggr`
            num_to_be_merged = int(min(embeddings_per_chunk, emb_part.shape[0]) - chunk_cluster_count)
            min_count_per_cluster = self.get_div_ceil_count(
                numer=chunk_cluster_count, denomin=len(torch.unique(Y_part))
            )

            # We want only one embedding vector for each cluster, so we calculate the number of embedding vectors to be removed
            class_target_vol = get_merge_quantity(
                num_to_be_removed=num_to_be_merged,
                pre_clus_labels=Y_part,
                min_count_per_cluster=min_count_per_cluster,
            )
            if not torch.jit.is_scripting():
                pbar.update(1)

            # `class_target_vol` is a list of cluster-indices from overclustering
            for spk_idx, merge_quantity in enumerate(list(class_target_vol)):
                merged_embs, merged_clus_labels, index_mapping = run_reducer(
                    pre_embs=emb_part, target_spk_idx=spk_idx, merge_quantity=merge_quantity, pre_clus_labels=Y_part,
                )
                total_emb.append(merged_embs)
                absolute_index_mapping = [x + offset_index for x in index_mapping]
                absolute_merge_mapping.append(absolute_index_mapping)
                window_range_list.append([window_offset, window_offset + merged_embs.shape[0]])
                window_offset += merged_embs.shape[0]

        if not torch.jit.is_scripting():
            pbar.close()

        # Concatenate the reduced embeddings then perform high-level clustering
        reduced_embs = torch.cat(total_emb)
        reduced_mat = getCosAffinityMatrix(reduced_embs)

        # Step-4: Map the aggregated labels `Y_aggr` back to the original labels for all `org_len` input embeddings: `Y_unpack`
        Y_aggr = self.speaker_clustering.forward_unit_infer(
            mat=reduced_mat,
            oracle_num_speakers=oracle_num_speakers,
            max_rp_threshold=max_rp_threshold,
            max_num_speakers=max_num_speakers,
            sparse_search_volume=sparse_search_volume,
            fixed_thres=fixed_thres,
        )
        if reduced_embs.shape[0] != Y_aggr.shape[0]:
            raise ValueError(
                f"The number of embeddings ({reduced_embs.shape[0]}) and the number of clustered labels ({Y_aggr.shape[0]}) do not match."
            )

        # Reassign the labels to the original embeddings
        Y_unpack = self.unpack_labels(
            Y_aggr=Y_aggr,
            window_range_list=window_range_list,
            absolute_merge_mapping=absolute_merge_mapping,
            org_len=emb.shape[0],
        )
        if Y_unpack.shape[0] != emb.shape[0]:
            raise ValueError(
                f"The number of raw input embeddings ({emb.shape[0]}) and the number of clustered labels ({Y_unpack.shape[0]}) do not match."
            )
        return Y_unpack

    def forward_infer(
        self,
        embeddings_in_scales: torch.Tensor,
        timestamps_in_scales: torch.Tensor,
        multiscale_segment_counts: torch.LongTensor,
        multiscale_weights: torch.Tensor,
        oracle_num_speakers: int = -1,
        max_rp_threshold: float = 0.15,
        max_num_speakers: int = 8,
        enhanced_count_thres: int = 80,
        sparse_search_volume: int = 30,
        fixed_thres: float = -1.0,
        chunk_cluster_count: int = 50,
        embeddings_per_chunk: int = 10000,
    ) -> torch.LongTensor:
        """
        This function is a wrapper designed for toggling between long-form and short-form speaker clustering.
        The details of short-form clustering is in `SpeakerClustering` class.
        NOTE: `torch.jit.script` currently does not support `**kwargs` in the function signature therefore,
        we need to use a wrapper function to handle the arguments.
        """
        if embeddings_per_chunk is not None and torch.max(multiscale_segment_counts) > embeddings_per_chunk:
            return self.long_forward_infer(
                embeddings_in_scales=embeddings_in_scales,
                timestamps_in_scales=timestamps_in_scales,
                multiscale_segment_counts=multiscale_segment_counts,
                multiscale_weights=multiscale_weights,
                oracle_num_speakers=oracle_num_speakers,
                max_rp_threshold=max_rp_threshold,
                max_num_speakers=max_num_speakers,
                sparse_search_volume=sparse_search_volume,
                fixed_thres=fixed_thres,
                chunk_cluster_count=chunk_cluster_count,
                embeddings_per_chunk=embeddings_per_chunk,
            )
        else:
            cluster_labels = self.speaker_clustering.forward_infer(
                embeddings_in_scales=embeddings_in_scales,
                timestamps_in_scales=timestamps_in_scales,
                multiscale_segment_counts=multiscale_segment_counts,
                multiscale_weights=multiscale_weights,
                oracle_num_speakers=oracle_num_speakers,
                max_rp_threshold=max_rp_threshold,
                max_num_speakers=max_num_speakers,
                enhanced_count_thres=enhanced_count_thres,
                sparse_search_volume=sparse_search_volume,
                fixed_thres=fixed_thres,
            )
            self.timestamps_in_scales = self.speaker_clustering.timestamps_in_scales
            return cluster_labels



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from inspect import isclass
from typing import Any, Type

import pytest
import torch
import torch.nn as nn

import nemo.core.neural_types.elements as nelements
from nemo.core.classes import Exportable, NeuralModule, typecheck
from nemo.core.classes.mixins import AdapterModuleMixin
from nemo.core.neural_types import FloatType, NeuralType, VoidType


def get_all_neural_types() -> list[Type[nelements.ElementType]]:
    """Get all neural types (elements) by inspecting neural_types.element module"""
    neural_types = []
    for neural_type_str in dir(nelements):
        candidate = getattr(nelements, neural_type_str)
        if (
            isclass(candidate)
            and issubclass(candidate, nelements.ElementType)
            and candidate is not nelements.ElementType
        ):
            neural_types.append(candidate)
    return neural_types


class SimpleLinear(NeuralModule):
    """Simple linear projection. Test use of NeuralModule instead of nn.Module"""

    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(10, 20, bias=False)

    def forward(self, x):
        return self.linear(x)


class SimpleLinearExportable(NeuralModule, Exportable):
    """Simple linear projection. Test use of NeuralModule with Exportable mixin"""

    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(10, 20, bias=False)

    def forward(self, x):
        return self.linear(x)


class SimpleLinearWithAdapterMixin(NeuralModule, AdapterModuleMixin):
    """Simple linear projection. Test use of NeuralModule with AdapterModuleMixin mixin"""

    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(10, 20, bias=False)

    def forward(self, x):
        return self.linear(x)


class SimpleLinearWithTypes(NeuralModule):
    """Simple linear projection. Test use of NeuralModule with input/output types"""

    @property
    def input_types(self) -> dict[str, Any]:
        return {
            "x": NeuralType(None, FloatType()),
        }

    @property
    def output_types(self) -> dict[str, Any]:
        return {
            "output": NeuralType(None, VoidType()),
        }

    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(10, 20, bias=False)

    @typecheck()
    def forward(self, x):
        return self.linear(x)


class DummyModuleWithIOTypes(NeuralModule):
    """Identity module. For testing input/output types in a sequence of network calls"""

    @property
    def input_types(self) -> dict[str, NeuralType]:
        return {
            "x": NeuralType(None, VoidType()),
        }

    @property
    def output_types(self) -> dict[str, NeuralType]:
        return {
            "output": NeuralType(None, FloatType()),
        }

    def forward(self, x):
        return x


class TestTorchJitCompatibility:
    @pytest.mark.unit
    def test_simple_linear(self):
        """Test basic module derived from NeuralModule"""
        module = torch.jit.script(SimpleLinear())
        x = torch.zeros(2, 10)
        result = module(x)
        assert result.shape == (2, 20)
        assert torch.allclose(result, torch.zeros_like(result))

    @pytest.mark.unit
    def test_simple_linear_exportable(self):
        """Test basic module derived from NeuralModule and Exportable"""
        module = torch.jit.script(SimpleLinearExportable())
        x = torch.zeros(2, 10)
        result = module(x)
        assert result.shape == (2, 20)
        assert torch.allclose(result, torch.zeros_like(result))

    @pytest.mark.unit
    def test_simple_linear_with_adapter_mixin(self):
        """Test basic module derived from NeuralModule and Adapter Mixin"""
        module = torch.jit.script(SimpleLinearWithAdapterMixin())
        x = torch.zeros(2, 10)
        result = module(x)
        assert result.shape == (2, 20)
        assert torch.allclose(result, torch.zeros_like(result))

    @pytest.mark.unit
    def test_simple_linear_with_types(self):
        """Test basic module derived from NeuralModule containing types"""
        module = torch.jit.script(SimpleLinearWithTypes())
        x = torch.zeros(2, 10)
        result = module(x)
        assert result.shape == (2, 20)
        assert torch.allclose(result, torch.zeros_like(result))

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "neural_type", get_all_neural_types(),
    )
    def test_element_compilable(self, neural_type: Type[nelements.ElementType]):
        """
        Tests that all NeuralType Elements are compilable by TorchScript (and be used in modules/functions).
        """

        @torch.jit.script
        def identity(x: torch.Tensor):
            if isinstance(neural_type(), nelements.VoidType):
                return x
            else:
                return x + 1

        _ = identity(torch.tensor(1.0))

    @pytest.mark.unit
    def test_dummy_module_with_io_types(self):
        """Test module with input/output types"""
        module = torch.jit.script(DummyModuleWithIOTypes())
        x = torch.rand(2, 10)
        result = module(x)
        assert result.shape == x.shape
        assert torch.allclose(result, x)

    @pytest.mark.unit
    def test_chain_with_types(self):
        """Test applying 2 modules consecutively with input/output types"""
        dummy_module = torch.jit.script(DummyModuleWithIOTypes())
        module = torch.jit.script(SimpleLinearWithTypes())
        x = torch.zeros(2, 10)
        result = module(dummy_module(x))
        assert result.shape == (2, 20)
        assert torch.allclose(result, torch.zeros_like(result))
